# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Package for objects which are related with Icon Services"""
from abc import ABCMeta, abstractmethod, ABC
from functools import wraps
from inspect import isfunction
from typing import List, Union, Optional, Dict

from iconcommons.logger import Logger
from typing_extensions import TypedDict

from .base.address import Address, AddressPrefix, SYSTEM_SCORE_ADDRESS, ZERO_SCORE_ADDRESS
from .base.exception import IconScoreException
from .icon_constant import IconServiceFlag
from .iconscore.icon_container_db import VarDB, DictDB, ArrayDB
from .iconscore.icon_score_base import interface, eventlog, external, payable, IconScoreBase, IconScoreDatabase
from .iconscore.icon_score_base2 import (
    InterfaceScore, revert, sha3_256, sha_256, json_loads, json_dumps,
    get_main_prep_info, get_sub_prep_info, recover_key,
    create_address_with_key, create_interface_score
)

from .iconscore.icon_system_score_base import IconSystemScoreBase
from .iconscore.system_score import InterfaceSystemScore
from .__version__ import __version__



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from enum import Flag
from typing import TYPE_CHECKING, Optional

from iconcommons import Logger
from iconservice.database.db import KeyValueDatabase
from iconservice.database.wal import WriteAheadLogWriter
from iconservice.icon_constant import ROLLBACK_LOG_TAG
from iconservice.rollback import get_backup_filename

if TYPE_CHECKING:
    from iconservice.database.wal import IissWAL
    from iconservice.base.block import Block
    from iconservice.database.batch import BlockBatch

TAG = ROLLBACK_LOG_TAG


class WALBackupState(Flag):
    CALC_PERIOD_END_BLOCK = 1


class BackupManager(object):
    """Backup and rollback for the previous block state

    """

    def __init__(self, backup_root_path: str, rc_data_path: str):
        Logger.debug(tag=TAG,
                     msg=f"__init__() start: "
                         f"backup_root_path={backup_root_path}, "
                         f"rc_data_path={rc_data_path}")

        self._rc_data_path = rc_data_path
        self._backup_root_path = backup_root_path

        Logger.info(tag=TAG, msg=f"backup_root_path={self._backup_root_path}")
        Logger.debug(tag=TAG, msg="__init__() end")

    def _get_backup_file_path(self, block_height: int) -> str:
        """

        :param block_height: the height of block to rollback to
        :return: backup state file path
        """
        assert block_height >= 0

        filename = get_backup_filename(block_height)
        return os.path.join(self._backup_root_path, filename)

    def run(self,
            icx_db: 'KeyValueDatabase',
            rc_db: 'KeyValueDatabase',
            revision: int,
            prev_block: 'Block',
            block_batch: 'BlockBatch',
            iiss_wal: 'IissWAL',
            is_calc_period_start_block: bool,
            instant_block_hash: bytes):
        """Backup the previous block state

        :param icx_db:
        :param rc_db:
        :param revision:
        :param prev_block: the latest confirmed block height during commit
        :param block_batch:
        :param iiss_wal:
        :param is_calc_period_start_block:
        :param instant_block_hash:
        :return:
        """
        Logger.debug(tag=TAG, msg="backup() start")

        path: str = self._get_backup_file_path(prev_block.height)
        Logger.info(tag=TAG, msg=f"backup_file_path={path}")

        writer = WriteAheadLogWriter(
            revision, max_log_count=2, block=prev_block, instant_block_hash=instant_block_hash)
        writer.open(path)

        if is_calc_period_start_block:
            writer.write_state(WALBackupState.CALC_PERIOD_END_BLOCK.value)

        self._backup_rc_db(writer, rc_db, iiss_wal)
        self._backup_state_db(writer, icx_db, block_batch)

        writer.close()

        Logger.debug(tag=TAG, msg="backup() end")

    @classmethod
    def _backup_rc_db(cls, writer: 'WriteAheadLogWriter', db: 'KeyValueDatabase', iiss_wal: 'IissWAL'):
        def get_rc_db_generator():
            for key, _ in iiss_wal:
                value: Optional[bytes] = db.get(key)
                yield key, value

        writer.write_walogable(get_rc_db_generator())

    @classmethod
    def _backup_state_db(cls, writer: 'WriteAheadLogWriter', db: 'KeyValueDatabase', block_batch: 'BlockBatch'):
        if block_batch is None:
            block_batch = {}

        def get_state_db_generator():
            for key in block_batch:
                value: Optional[bytes] = db.get(key)
                yield key, value

        writer.write_walogable(get_state_db_generator())



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from inspect import signature, Signature, Parameter, isclass, getmembers, isfunction
from typing import Any, Optional, TYPE_CHECKING

from .icon_score_constant import ScoreFlag, CONST_INDEXED_ARGS_COUNT, BaseType, \
    STR_FALLBACK, STR_ON_INSTALL, STR_ON_UPDATE
from .typing.element import get_score_flag, is_any_score_flag_on
from ..base.address import Address
from ..base.exception import IllegalFormatException, InvalidParamsException
from ..base.type_converter import TypeConverter
from ..icon_constant import Revision
from ..utils import get_main_type_from_annotations_type

if TYPE_CHECKING:
    from .icon_score_base import IconScoreBase
    from .icon_score_context import IconScoreContext


class ScoreApiGenerator:

    __API_TYPE = 'type'
    __API_NAME = 'name'
    __API_INPUTS = 'inputs'
    __API_OUTPUTS = 'outputs'
    __API_PAYABLE = 'payable'
    __API_READONLY = 'readonly'
    __API_INPUTS_INDEXED = 'indexed'
    __API_INPUTS_DEFAULT = 'default'
    __API_PARAMS_ADDRESS = 'Address'
    __API_PARAMS_INDEXED = 'Indexed'
    __API_TYPE_FUNCTION = 'function'
    __API_TYPE_EVENT = 'eventlog'

    __API_TYPE_FALLBACK = STR_FALLBACK
    __API_TYPE_ON_INSTALL = STR_ON_INSTALL
    __API_TYPE_ON_UPDATE = STR_ON_UPDATE

    __on_deploy = [__API_TYPE_ON_INSTALL, __API_TYPE_ON_UPDATE]

    @staticmethod
    def generate(score_funcs: list) -> list:
        api = []
        ScoreApiGenerator.__generate_functions(api, score_funcs)
        ScoreApiGenerator.__generate_events(api, score_funcs)
        return api

    @staticmethod
    def check_on_deploy(context: 'IconScoreContext', score: 'IconScoreBase') -> None:
        custom_funcs = [value for key, value in getmembers(score.__class__, predicate=isfunction)
                        if key in ScoreApiGenerator.__on_deploy]
        for func in custom_funcs:
            ScoreApiGenerator.__check_on_deploy_function(context, signature(func))

    @staticmethod
    def __check_on_deploy_function(context: 'IconScoreContext', sig_info: 'Signature') -> None:
        params = dict(sig_info.parameters)
        for param_name, param in params.items():
            if param_name == 'self' or param_name == 'cls':
                continue
            if context.revision > Revision.TWO.value or param.kind != Parameter.VAR_KEYWORD:
                ScoreApiGenerator.__generate_input([], param, False)

    @staticmethod
    def __generate_functions(src: list, score_funcs: list) -> None:
        for func in score_funcs:
            score_flag = get_score_flag(func)
            is_readonly = bool(score_flag & ScoreFlag.READONLY)
            is_payable = bool(score_flag & ScoreFlag.PAYABLE)

            try:
                if score_flag & ScoreFlag.EXTERNAL:
                    src.append(ScoreApiGenerator.__generate_normal_function(
                        func.__name__, is_readonly, is_payable, signature(func)))
                elif func.__name__ == ScoreApiGenerator.__API_TYPE_FALLBACK:
                    if is_payable:
                        src.append(ScoreApiGenerator.__generate_fallback_function(
                            func.__name__, is_payable, signature(func)))
            except IllegalFormatException as e:
                raise IllegalFormatException(f"{e.message} at {func.__name__}")

    @staticmethod
    def __generate_normal_function(func_name: str, is_readonly: bool, is_payable: bool, sig_info: 'Signature') -> dict:
        info = dict()
        info[ScoreApiGenerator.__API_TYPE] = \
            ScoreApiGenerator.__API_TYPE_FUNCTION
        info[ScoreApiGenerator.__API_NAME] = func_name
        info[ScoreApiGenerator.__API_INPUTS] = \
            ScoreApiGenerator.__generate_inputs(dict(sig_info.parameters))
        info[ScoreApiGenerator.__API_OUTPUTS] = \
            ScoreApiGenerator.__generate_output(
                sig_info.return_annotation, is_readonly)

        if is_readonly:
            info[ScoreApiGenerator.__API_READONLY] = is_readonly
        if is_payable:
            info[ScoreApiGenerator.__API_PAYABLE] = is_payable

        return info

    @staticmethod
    def __generate_fallback_function(func_name: str, is_payable: bool, sig_info: 'Signature') -> dict:
        info = dict()
        info[ScoreApiGenerator.__API_TYPE] = ScoreApiGenerator.__API_TYPE_FALLBACK
        info[ScoreApiGenerator.__API_NAME] = func_name

        if len(sig_info.parameters) > 1:
            raise InvalidParamsException("Invalid fallback signature")

        if sig_info.return_annotation is not None:
            if sig_info.return_annotation is not Signature.empty:
                raise InvalidParamsException("Invalid fallback signature")

        info[ScoreApiGenerator.__API_PAYABLE] = is_payable

        return info

    @staticmethod
    def __generate_events(src: list, score_funcs: list) -> None:
        event_funcs = {
            func.__name__: signature(func) for func in score_funcs
            if is_any_score_flag_on(func, ScoreFlag.EVENTLOG)
        }

        indexed_args_counts = {
            func.__name__: getattr(func, CONST_INDEXED_ARGS_COUNT, 0)
            for func in score_funcs
            if getattr(func, CONST_INDEXED_ARGS_COUNT, 0)
        }

        for func_name, event in event_funcs.items():
            index_args_count = indexed_args_counts.get(func_name, 0)
            src.append(ScoreApiGenerator.__generate_event(func_name, event, index_args_count))

    @staticmethod
    def __generate_event(func_name: str, sig_info: 'Signature', index_args_count: int) -> dict:
        info = dict()
        info[ScoreApiGenerator.__API_TYPE] = ScoreApiGenerator.__API_TYPE_EVENT
        info[ScoreApiGenerator.__API_NAME] = func_name
        info[ScoreApiGenerator.__API_INPUTS] = \
            ScoreApiGenerator.__generate_inputs(dict(sig_info.parameters), index_args_count)
        return info

    @staticmethod
    def __generate_output(params_type: Any, is_readonly: bool) -> list:
        info_list = []

        if not is_readonly:
            return info_list

        if params_type is Signature.empty:
            raise IllegalFormatException(
                "Returning type should be declared in read-only functions")

        main_type = get_main_type_from_annotations_type(params_type)
        main_type = ScoreApiGenerator.__convert_str_to_type(main_type)

        # At first, finds if the type is a 'list' or a 'dict'
        # if not, finds a base type
        find = (t for t in [list, dict]
                if isclass(main_type) and issubclass(main_type, t))
        api_type = next(find, None)
        if api_type is None:
            api_type = ScoreApiGenerator.__find_base_super_type(main_type)
        if api_type is None:
            raise IllegalFormatException(f"Unsupported type for '{params_type}'")

        info = dict()
        info[ScoreApiGenerator.__API_TYPE] = api_type.__name__
        info_list.append(info)
        return info_list

    @staticmethod
    def __convert_str_to_type(params_type: Any) -> Any:
        if not isinstance(params_type, str):
            return params_type

        if params_type == 'Address':
            return Address
        else:
            return params_type

    @staticmethod
    def __generate_inputs(params: dict, index_args_count: int = 0) -> list:
        tmp_list = []
        args_index = 0
        for param_name, param in params.items():
            if param_name == 'self' or param_name == 'cls':
                continue
            is_indexed = args_index < index_args_count
            args_index += 1
            ScoreApiGenerator.__generate_input(tmp_list, param, is_indexed)
        return tmp_list

    @staticmethod
    def __generate_input(src: list, param: 'Parameter', is_indexed: bool):
        # If there's no hint of argument in the function declaration,
        # raise an exception
        if param.annotation is Parameter.empty:
            raise IllegalFormatException(f"Missing argument hint for '{param.name}'")

        main_type = get_main_type_from_annotations_type(param.annotation)
        main_type = ScoreApiGenerator.__convert_str_to_type(main_type)
        api_type = ScoreApiGenerator.__find_base_super_type(main_type)
        if api_type is None:
            raise IllegalFormatException(
                f"Unsupported type for '{param.name}: {param.annotation}'")
        info = dict()
        info[ScoreApiGenerator.__API_NAME] = param.name
        info[ScoreApiGenerator.__API_TYPE] = api_type.__name__
        if is_indexed:
            info[ScoreApiGenerator.__API_INPUTS_INDEXED] = is_indexed
        if param.default is not Parameter.empty:
            if param.default is not None and not isinstance(param.default, main_type):
                raise InvalidParamsException(f'Default params type mismatch. value: {param.default} type: {main_type}')
            info[ScoreApiGenerator.__API_INPUTS_DEFAULT] = TypeConverter.convert_type_reverse(param.default)
        src.append(info)

    @staticmethod
    def __find_base_super_type(t: type) -> Optional[type]:
        """
        Finds a base type of the input and returns it if any
        :param t: target
        :return: base_super_type
        """
        find = (base_type for base_type in BaseType.__constraints__
                if isclass(t) and issubclass(t, base_type))
        return next(find, None)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Flag, unique
from typing import TypeVar

from ..base.address import Address

T = TypeVar('T')
# TODO add checking function for list, dict type
BaseType = TypeVar("BaseType", bool, int, str, bytes, list, dict, Address)

CONST_CLASS_API = '__api'
CONST_CLASS_ELEMENT_METADATAS = '__element_metadatas'

CONST_SCORE_FLAG = '__score_flag'
CONST_INDEXED_ARGS_COUNT = '__indexed_args_count'

FORMAT_IS_NOT_FUNCTION_OBJECT = "isn't function object: {}, cls: {}"
FORMAT_IS_NOT_DERIVED_OF_OBJECT = "isn't derived of {}"
FORMAT_DECORATOR_DUPLICATED = "can't duplicated {} decorator func: {}, cls: {}"

STR_FALLBACK = 'fallback'
STR_ON_INSTALL = 'on_install'
STR_ON_UPDATE = 'on_update'

ATTR_SCORE_GET_API = "_IconScoreBase__get_api"
ATTR_SCORE_CALL = "_IconScoreBase__call"
ATTR_SCORE_VALIDATE_EXTERNAL_METHOD = "_IconScoreBase__validate_external_method"


@unique
class ScoreFlag(Flag):
    NONE = 0

    # Used for external function
    READONLY = 0x01
    EXTERNAL = 0x02
    PAYABLE = 0x04
    FALLBACK = 0x08
    FUNC = 0xFF

    # Used for eventlog declaration in score
    EVENTLOG = 0x100

    # Used for interface declaration in score
    INTERFACE = 0x10000

    ALL = 0xFFFFFF



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hashlib import sha3_256
from typing import TYPE_CHECKING

from .deposit import Deposit
from .deposit_meta import DepositMeta
from ..base.ComponentBase import StorageBase
from ..base.address import Address

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext


class Storage(StorageBase):
    """Fee and Deposit state manager embedding a state db wrapper"""

    _FEE_PREFIX = b'\x02'

    def _generate_key(self, key_data: bytes):
        """
        Generates a db key
        """
        return self._FEE_PREFIX + sha3_256(key_data).digest()

    def get_deposit_meta(self, context: 'IconScoreContext', score_address: 'Address') -> 'DepositMeta':
        """Returns the score deposit meta information.

        :param context: Object that contains the useful information to process user's JSON-RPC request
        :param score_address: SCORE address
        :return: DepositMeta object
        """
        key = self._generate_key(score_address.to_bytes())
        value = self._db.get(context, key)
        return DepositMeta.from_bytes(value) if value else None

    def put_deposit_meta(self, context: 'IconScoreContext', score_address: 'Address', deposit_meta: 'DepositMeta'):
        """Puts the score deposit meta information into db.

        :param context: Object that contains the useful information to process user's JSON-RPC request
        :param score_address: SCORE address
        :param deposit_meta: DepositMeta object
        :return: None
        """
        key = self._generate_key(score_address.to_bytes())
        value = deposit_meta.to_bytes()
        self._db.put(context, key, value)

    def delete_deposit_meta(self, context: 'IconScoreContext', score_address: 'Address'):
        """Deletes the score deposit meta information from db.

        :param context: Object that contains the useful information to process user's JSON-RPC request
        :param score_address: SCORE address
        :return: None
        """
        key = self._generate_key(score_address.to_bytes())
        self._db.delete(context, key)

    def get_deposit(self, context: 'IconScoreContext', deposit_id: bytes) -> 'Deposit':
        """Returns the deposit.

        :param context: Object that contains the useful information to process user's JSON-RPC request
        :param deposit_id: Deposit id
        :return: Deposit Object
        """
        key = self._generate_key(deposit_id)
        value = self._db.get(context, key)

        if value:
            value = Deposit.from_bytes(value)
            value.id = deposit_id

        return value

    def put_deposit(self, context: 'IconScoreContext', deposit: 'Deposit'):
        """Puts the deposit data into db.

        :param context: Object that contains the useful information to process user's JSON-RPC request
        :param deposit: Deposit Object
        :return: None
        """
        key = self._generate_key(deposit.id)
        value = deposit.to_bytes()
        self._db.put(context, key, value)

    def delete_deposit(self, context: 'IconScoreContext', deposit_id: bytes):
        """Deletes the deposit from db.

        :param context: Object that contains the useful information to process user's JSON-RPC request
        :param deposit_id: Deposit id
        :return: None
        """
        key = self._generate_key(deposit_id)
        self._db.delete(context, key)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from enum import IntEnum, IntFlag
from typing import TYPE_CHECKING, Optional, Union

from iconcommons import Logger

from .coin_part import CoinPart, CoinPartFlag, CoinPartType
from .delegation_part import DelegationPart
from .icx_account import Account
from .stake_part import StakePart
from ..base.ComponentBase import StorageBase
from ..base.address import Address
from ..base.block import Block, NULL_BLOCK
from ..icon_constant import DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER, ICX_LOG_TAG, ROLLBACK_LOG_TAG, IconScoreContextType
from ..utils import bytes_to_hex

if TYPE_CHECKING:
    from ..database.db import ContextDatabase
    from ..iconscore.icon_score_context import IconScoreContext


class AccountPartFlag(IntFlag):
    """PartFlag Type
    """
    NONE = 0
    COIN = 1
    STAKE = 2
    DELEGATION = 4


class Intent(IntEnum):
    TRANSFER = AccountPartFlag.COIN
    STAKE = AccountPartFlag.COIN | AccountPartFlag.STAKE
    DELEGATED = AccountPartFlag.DELEGATION
    INIT_PREP = AccountPartFlag.STAKE | AccountPartFlag.DELEGATION
    ALL = AccountPartFlag.COIN | AccountPartFlag.STAKE | AccountPartFlag.DELEGATION


class Storage(StorageBase):
    """Icx coin state manager embedding a state db wrapper"""

    _GENESIS_DB_KEY = 'genesis'
    _TREASURY_DB_KEY = 'fee_treasury'

    # Level db keys
    LAST_BLOCK_KEY = b'last_block'
    _TOTAL_SUPPLY_KEY = b'total_supply'

    def __init__(self, db: 'ContextDatabase'):
        """Constructor

        :param db: (Database) state db wrapper
        """
        super().__init__(db)
        self._db = db
        self._last_block = NULL_BLOCK
        self._genesis: Optional['Address'] = None
        self._fee_treasury: Optional['Address'] = None

    def open(self, context: 'IconScoreContext'):
        self._load_special_address(context, self._GENESIS_DB_KEY)
        self._load_special_address(context, self._TREASURY_DB_KEY)

    def rollback(self, context: 'IconScoreContext', block_height: int, block_hash: bytes):
        Logger.info(tag=ROLLBACK_LOG_TAG,
                    msg=f"rollback() start: block_height={block_height} block_hash={bytes_to_hex(block_hash)}")

        self._load_special_address(context, self._GENESIS_DB_KEY)
        self._load_special_address(context, self._TREASURY_DB_KEY)
        self.load_last_block_info(context)

        Logger.info(tag=ROLLBACK_LOG_TAG, msg="rollback() end")

    @property
    def last_block(self) -> 'Block':
        return self._last_block

    @property
    def genesis(self) -> 'Address':
        return self._genesis

    @property
    def fee_treasury(self) -> 'Address':
        return self._fee_treasury

    def load_last_block_info(self, context: 'IconScoreContext'):
        block_bytes = self._db.get(context, self.LAST_BLOCK_KEY)
        if block_bytes is None:
            return

        self._last_block = Block.from_bytes(block_bytes)

    def _load_special_address(self,
                              context: 'IconScoreContext',
                              db_key: str):
        """Load address info from state db according to db_key

        :param context:
        :param db_key: db key info
        """
        Logger.debug(f'_load_address_from_storage() start(address type: {db_key})', ICX_LOG_TAG)
        text = context.storage.icx.get_text(context, db_key)
        if text:
            obj = json.loads(text)

            # Support to load MainNet 1.0 db
            address: str = obj['address']
            if len(address) == 40:
                address = f'hx{address}'

            address: Address = Address.from_string(address)
            if db_key == self._GENESIS_DB_KEY:
                self._genesis: 'Address' = address
            elif db_key == self._TREASURY_DB_KEY:
                self._fee_treasury: 'Address' = address
            Logger.info(f'{db_key}: {address}', ICX_LOG_TAG)
        Logger.debug(f'_load_address_from_storage() end(address type: {db_key})', ICX_LOG_TAG)

    def set_last_block(self, block: 'Block'):
        self._last_block = block

    def put_genesis_accounts(self, context: 'IconScoreContext', accounts: list):
        genesis = accounts[0]
        treasury = accounts[1]
        others = accounts[2:]

        __ADDRESS_KEY = 'address'
        __AMOUNT_KEY = 'balance'

        self._put_genesis_data_account(
            context=context,
            coin_part_type=CoinPartType.GENESIS,
            address=genesis[__ADDRESS_KEY],
            amount=genesis[__AMOUNT_KEY])

        self._put_genesis_data_account(
            context=context,
            coin_part_type=CoinPartType.TREASURY,
            address=treasury[__ADDRESS_KEY],
            amount=treasury[__AMOUNT_KEY])

        for other in others:
            self._put_genesis_data_account(
                context=context,
                coin_part_type=CoinPartType.GENERAL,
                address=other[__ADDRESS_KEY],
                amount=other[__AMOUNT_KEY])

    def _put_genesis_data_account(self,
                                  context: 'IconScoreContext',
                                  coin_part_type: 'CoinPartType',
                                  address: 'Address',
                                  amount: int):
        """This method is called only on invoking the genesis block

        :param context:
        :param coin_part_type:
        :param address:
        :param amount:
        :return:
        """

        coin_part: 'CoinPart' = CoinPart(coin_part_type)
        account: 'Account' = Account(address, context.block.height, context.revision, coin_part=coin_part)
        account.deposit(int(amount))
        if not account.coin_part.is_dirty():
            account.coin_part.set_dirty(True)
        self.put_account(context, account)

        if account.balance > 0:
            total_supply = self.get_total_supply(context)
            total_supply += account.balance
            context.storage.icx.put_total_supply(context, total_supply)
        if coin_part_type in [CoinPartType.GENESIS, CoinPartType.TREASURY]:
            self._put_special_account(context, account)

    def _put_special_account(self,
                             context: 'IconScoreContext',
                             account: 'Account'):
        """Compared to other general accounts,
        additional tasks should be processed
        for special accounts (genesis, treasury)

        :param context:
        :param account: genesis or treasury accounts
        """

        assert account.coin_part is not None
        assert account.coin_part.type in (CoinPartType.GENESIS, CoinPartType.TREASURY)

        if account.coin_part.type == CoinPartType.GENESIS:
            db_key = self._GENESIS_DB_KEY
            self._genesis = account.address
        else:
            db_key = self._TREASURY_DB_KEY
            self._fee_treasury = account.address

        obj = {'version': 0, 'address': str(account.address)}
        text = json.dumps(obj)

        context.storage.icx.put_text(context, db_key, text)

    def get_text(self, context: 'IconScoreContext', name: str) -> Optional[str]:
        """Returns text format value from db

        :return: (str or None)
            text value mapped by name
            default encoding: utf8
        """
        key = name.encode()
        value = self._db.get(context, key)
        if value:
            return value.decode()
        else:
            return None

    def put_text(self,
                 context: 'IconScoreContext',
                 name: str,
                 text: str):
        """Saves text to db with name as a key
        All text are utf8 encoded.

        :param context:
        :param name: db key
        :param text: db value
        """
        key = name.encode()
        value = text.encode()
        self._db.put(context, key, value)

    def get_account(self,
                    context: 'IconScoreContext',
                    address: 'Address',
                    intent: 'Intent' = Intent.TRANSFER) -> 'Account':

        """Returns the account indicated by address.

        :param context:
        :param address: account address
        :param intent:
        :return: (Account)
            If the account indicated by address is not present,
            create a new account.
        """

        coin_part: Optional['CoinPart'] = None
        stake_part: Optional['StakePart'] = None
        delegation_part: Optional['DelegationPart'] = None

        part_flags: 'AccountPartFlag' = AccountPartFlag(intent)

        if AccountPartFlag.COIN in part_flags:
            coin_part: 'CoinPart' = self._get_part(context, CoinPart, address)

            if CoinPartFlag.HAS_UNSTAKE in coin_part.flags:
                part_flags |= AccountPartFlag.STAKE

        if AccountPartFlag.STAKE in part_flags:
            stake_part: 'StakePart' = self._get_part(context, StakePart, address)

        if AccountPartFlag.DELEGATION in part_flags:
            delegation_part: 'DelegationPart' = self._get_part(context, DelegationPart, address)

        return Account(address, context.block.height, context.revision,
                       coin_part=coin_part,
                       stake_part=stake_part,
                       delegation_part=delegation_part)

    def get_treasury_account(self, context: 'IconScoreContext') -> 'Account':
        """Returns the instance of treasury account

        :param context:
        :return: Account
        """
        return context.storage.icx.get_account(context, context.storage.icx.fee_treasury)

    def _get_part(self, context: 'IconScoreContext',
                  part_class: Union[type(CoinPart), type(StakePart), type(DelegationPart)],
                  address: 'Address') -> Union['CoinPart', 'StakePart', 'DelegationPart']:
        key: bytes = part_class.make_key(address)
        value: bytes = self._db.get(context, key)

        if value is None and part_class is CoinPart:
            Logger.info(tag="PV", msg=f"No CoinPart: {address} {context.block}")

        return part_class.from_bytes(value) if value else part_class()

    def put_account(self,
                    context: 'IconScoreContext',
                    account: 'Account'):

        """Put account into to db.

        :param context:
        :param account: account to save
        """
        parts = [account.coin_part, account.stake_part, account.delegation_part]

        for part in parts:
            if part and part.is_dirty():
                key: bytes = part.make_key(account.address)

                if isinstance(part, CoinPart) or isinstance(part, StakePart):
                    value: bytes = part.to_bytes(context.revision)
                else:
                    value: bytes = part.to_bytes()

                self._db.put(context, key, value)

    def delete_account(self,
                       context: 'IconScoreContext',
                       account: 'Account'):
        """Delete account info from db.

        :param context:
        :param account: account to delete
        """
        raise Exception("not implemented")

    def is_address_present(self,
                           context: 'IconScoreContext',
                           address: 'Address') -> bool:
        """Checks whether value indicated by address is present or not.

        :param context:
        :param address: account address
        :return: True(present) False(not present)
        """
        key = address.to_bytes()
        value = self._db.get(context, key)

        return bool(value)

    def get_total_supply(self, context: 'IconScoreContext') -> int:
        """Returns the total supply.

        :return: (int) coin total supply in loop (1 icx == 1e18 loop)
        """
        value = self._db.get(context, self._TOTAL_SUPPLY_KEY)

        amount = 0
        if value:
            amount = int.from_bytes(value, DATA_BYTE_ORDER)

        return amount

    def put_total_supply(self,
                         context: 'IconScoreContext',
                         value: int):
        """Saves the total supply to db.

        :param context:
        :param value: coin total supply
        """
        value = value.to_bytes(DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER)
        self._db.put(context, self._TOTAL_SUPPLY_KEY, value)



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple

from .base_part import BasePart
from ..base.address import Address
from ..base.exception import InvalidParamsException
from ..utils.msgpack_for_db import MsgPackForDB


class DelegationPart(BasePart):
    _VERSION = 0
    PREFIX = b"aod|"

    def __init__(self, delegated_amount: int = 0, delegations: list = None):
        super().__init__()

        if delegations is None:
            delegations = []

        self._delegations: List[List['Address', int], ...] = delegations
        self._delegated_amount: int = delegated_amount
        self._delegations_amount: int = self._update_delegations_amount(delegations)

    def __str__(self):
        return f"delegations={self._delegations}, " \
               f"delegated_amount={self._delegated_amount}," \
               f"delegation_amount={self._delegations_amount}"

    @staticmethod
    def make_key(address: 'Address'):
        return DelegationPart.PREFIX + address.to_bytes_including_prefix()

    @property
    def delegated_amount(self) -> int:
        return self._delegated_amount

    @delegated_amount.setter
    def delegated_amount(self, value: int):
        if value < 0:
            raise InvalidParamsException(f"Invalid params: delegated_amount({value}) < 0")

        if value == self._delegated_amount:
            return
        self._delegated_amount = value
        self.set_dirty(True)

    @property
    def delegations(self) -> list:
        return self._delegations

    @property
    def delegations_amount(self) -> int:
        return self._delegations_amount

    @staticmethod
    def _update_delegations_amount(delegations: list) -> int:
        total_delegation_amount: int = 0

        for _, value in delegations:
            total_delegation_amount += value

        return total_delegation_amount

    def update_delegated_amount(self, offset: int):
        if offset == 0:
            return

        self._delegated_amount += offset

        if self._delegations_amount < 0:
            raise InvalidParamsException('Fail update_delegated_amount: delegations_amount < 0')
        
        self.set_dirty(True)

    def set_delegations(self, new_delegations: List[Tuple['Address', int]]):
        self._delegations: list = new_delegations
        self._delegations_amount: int = self._update_delegations_amount(new_delegations)

        self.set_dirty(True)

    @staticmethod
    def from_bytes(buf: bytes) -> 'DelegationPart':
        """Create DelegationPart object from bytes data

        :param buf: (bytes) bytes data including DelegationPart information
        :return: (DelegationPart) DelegationPart object
        """

        data: list = MsgPackForDB.loads(buf)
        version = data[0]

        assert version == DelegationPart._VERSION

        delegated_amount: int = data[1]
        delegation_list: list = data[2]
        delegations: list = []
        for i in range(0, len(delegation_list), 2):
            item = (delegation_list[i], delegation_list[i + 1])
            assert isinstance(item[0], Address)
            assert isinstance(item[1], int)

            delegations.append(item)

        return DelegationPart(delegated_amount=delegated_amount, delegations=delegations)

    def to_bytes(self) -> bytes:
        """Convert Account of Stake object to bytes

        :return: data including information of AccountOfDelegation object
        """

        data = [
            self._VERSION,
            self.delegated_amount
        ]
        delegations: list = []
        for address, value in self._delegations:
            delegations.append(address)
            delegations.append(value)
        data.append(delegations)

        return MsgPackForDB.dumps(data)

    def __eq__(self, other) -> bool:
        """operator == overriding

        :param other: (AccountOfDelegation)
        """

        return isinstance(other, DelegationPart) \
            and self._delegated_amount == other.delegated_amount \
            and self._delegations == other.delegations

    def __ne__(self, other) -> bool:
        """operator != overriding

        :param other: (AccountOfDelegation)
        """
        return not self.__eq__(other)



# -*- coding: utf-8 -*-
# Copyright 2020 ICON Foundation Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from enum import Flag, auto
from inspect import Signature, Parameter
from typing import Optional, Dict, Union, Any, List

from . import (
    BaseObject,
    is_base_type,
    is_struct,
    name_to_type,
    get_origin,
    get_args,
    get_annotations,
)
from .verification import set_default_value_to_params
from ...base.address import Address
from ...base.exception import InvalidParamsException


def base_object_to_str(value: Any) -> str:
    if isinstance(value, Address):
        return str(value)
    elif isinstance(value, int):
        return hex(value)
    elif isinstance(value, bytes):
        return bytes_to_hex(value)
    elif isinstance(value, bool):
        return hex(value)
    elif isinstance(value, str):
        return value

    raise TypeError(f"Unsupported type: {type(value)}")


def object_to_str(value: Any) -> Union[Any]:
    if is_base_type(type(value)):
        return base_object_to_str(value)

    if isinstance(value, list):
        return [object_to_str(i) for i in value]

    if isinstance(value, dict):
        return {k: object_to_str(value[k]) for k in value}

    if value is None:
        return None

    raise TypeError(f"Unsupported type: {type(value)}")


def str_to_base_object_by_type_name(type_name: str, value: str) -> BaseObject:
    return str_to_base_object(value, name_to_type(type_name))


def str_to_int(value: str) -> int:
    if isinstance(value, int):
        return value

    base = 16 if is_hex(value) else 10
    return int(value, base)


def str_to_base_object(value: str, type_hint: type) -> BaseObject:
    if not isinstance(value, str):
        raise InvalidParamsException(f"Type mismatch: value={value} type_hint={type_hint}")

    if type_hint is bool:
        return bool(str_to_int(value))
    if type_hint is bytes:
        return hex_to_bytes(value)
    if type_hint is int:
        return str_to_int(value)
    if type_hint is str:
        return value
    if type_hint is Address:
        return Address.from_string(value)

    raise InvalidParamsException(f"Unknown type: {type_hint}")


def bytes_to_hex(value: bytes, prefix: str = "0x") -> str:
    return f"{prefix}{value.hex()}"


def hex_to_bytes(value: Optional[str]) -> Optional[bytes]:
    if value is None:
        return None

    if value.startswith("0x"):
        value = value[2:]

    return bytes.fromhex(value)


def is_hex(value: str) -> bool:
    return value.startswith("0x") or value.startswith("-0x")


class ConvertOption(Flag):
    NONE = 0
    IGNORE_UNKNOWN_PARAMS = auto()


def convert_score_parameters(
        params: Dict[str, Any],
        sig: Signature,
        options: ConvertOption = ConvertOption.NONE):
    """Convert string values in score parameters to object values

    :param params:
    :param sig:
    :param options:
    :return:
    """
    _verify_arguments(params, sig)

    converted_params = {}
    parameters = sig.parameters

    for k, v in params.items():
        if not isinstance(k, str):
            raise InvalidParamsException(f"Invalid key type: key={k}")

        try:
            parameter: Parameter = parameters[k]
            converted_params[k] = str_to_object(v, parameter.annotation)
        except KeyError:
            if not (options & ConvertOption.IGNORE_UNKNOWN_PARAMS):
                raise InvalidParamsException(f"Unknown param: key={k} value={v}")

    set_default_value_to_params(params, parameters)

    return converted_params


def _verify_arguments(params: Dict[str, Any], sig: Signature):
    """Check if all required arguments are present

    :param params:
    :param sig: normalized signature
    :return:
    """
    parameters = sig.parameters

    for k in parameters:
        parameter: Parameter = parameters[k]
        param = params.get(k, parameter.default)

        if param is Parameter.empty:
            raise InvalidParamsException(f"Argument not found: {k}")


def str_to_object(value: Union[str, list, dict, None], type_hint: type) -> Any:
    if not isinstance(value, (dict, list, str, type(None))):
        raise InvalidParamsException(f"Invalid value type: {value}")

    origin = get_origin(type_hint)

    if is_base_type(origin):
        return str_to_base_object(value, origin)
    elif is_struct(origin):
        return str_to_object_in_struct(value, type_hint)
    elif origin is list:
        return str_to_object_in_list(value, type_hint)
    elif origin is dict:
        return str_to_object_in_dict(value, type_hint)
    elif origin is Union:
        # Assume that only the specific type of Union (= Optional) is allowed in iconservice
        return str_to_object_in_union(value, type_hint)

    raise InvalidParamsException(f"Type mismatch: value={value} type_hint={type_hint}")


def str_to_object_in_struct(value: Dict[str, Optional[str]], type_hint: type) -> Dict[str, Any]:
    if not isinstance(value, dict):
        raise InvalidParamsException(f"Type mismatch: value={value} type_hint={type_hint}")

    annotations = get_annotations(type_hint, None)

    ret = OrderedDict()

    for k, v in value.items():
        if k not in annotations:
            raise InvalidParamsException(f"Unknown field in struct: key={k}")

        ret[k] = str_to_object(v, annotations[k])

    if len(ret) != len(annotations):
        raise InvalidParamsException(f"Missing field in struct")

    return ret


def str_to_object_in_list(value: List[Any], type_hint: type) -> List[Any]:
    if not isinstance(value, list):
        raise InvalidParamsException(f"Type mismatch: value={value} type_hint={type_hint}")

    args = get_args(type_hint)
    return [str_to_object(i, args[0]) for i in value]


def str_to_object_in_dict(value: Dict[str, Any], type_hint: type) -> Dict[str, Any]:
    if not isinstance(value, dict):
        raise InvalidParamsException(f"Type mismatch: value={value} type_hint={type_hint}")

    args = get_args(type_hint)
    return OrderedDict(
        (k, str_to_object(v, type_hint=args[1])) for k, v in value.items()
    )


def str_to_object_in_union(value: Union[Any], type_hint: type) -> Optional[Any]:
    args = get_args(type_hint)
    return None if value is None else str_to_object(value, args[0])



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
from abc import ABC, ABCMeta
from enum import IntEnum
from typing import TYPE_CHECKING, Optional, Any, Callable
from typing import Tuple, List

from coincurve import PublicKey

from .icon_score_constant import FORMAT_IS_NOT_DERIVED_OF_OBJECT, T
from ..base.address import Address, AddressPrefix
from ..base.exception import InvalidParamsException, IconScoreException, InvalidInstanceException
from ..icon_constant import CHARSET_ENCODING
from ..icon_constant import Revision
from ..iconscore.context.context import ContextContainer
from ..iconscore.icon_score_step import StepType

if TYPE_CHECKING:
    from .icon_score_context import IconScoreContext


class InterfaceScoreMeta(ABCMeta):
    def __new__(mcs, name, bases, namespace, **kwargs):
        if ABC in bases:
            return super().__new__(mcs, name, bases, namespace, **kwargs)

        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        return cls


class InterfaceScore(ABC, metaclass=InterfaceScoreMeta):
    """
    An interface class that is used to invoke other SCORE’s external method.
    """
    def __init__(self, addr_to: 'Address'):
        """
        A Python init function. Invoked when the contract call create_interface_score()
        """
        self.__addr_to = addr_to
        self.__icx = 0

    @property
    def addr_to(self) -> 'Address':
        """
        The address of SCORE to invoke

        :return: :class:`.Address` SCORE address
        """
        return self.__addr_to

    def icx(self, value: int):
        if not (isinstance(value, int) and value >= 0):
            raise InvalidParamsException(f"Invalid icx: {value}")

        self.__icx = value
        return self

    def __get_icx(self) -> int:
        return self.__icx

    def __reset_icx(self):
        self.__icx = 0


class Block(object):
    def __init__(self, block_height: int, timestamp: int) -> None:
        """Constructor

        :param block_height: block height
        :param timestamp: block timestamp
        """
        self._height = block_height
        # unit: microsecond
        self._timestamp = timestamp

    @property
    def height(self) -> int:
        return self._height

    @property
    def timestamp(self) -> int:
        return self._timestamp


class ScoreApiStepRatio(IntEnum):
    SHA3_256 = 1000
    SHA_256 = 1000
    CREATE_ADDRESS_WITH_COMPRESSED_KEY = 15000
    CREATE_ADDRESS_WITH_UNCOMPRESSED_KEY = 1500
    JSON_DUMPS = 5000
    JSON_LOADS = 4000
    RECOVER_KEY = 70000


def _get_api_call_step_cost(context: 'IconScoreContext', ratio: ScoreApiStepRatio) -> int:
    """Returns the step cost for a given SCORE API

    API CALL step cost in context.step_counter means the step cost of sha3_256(b'')
    Each step cost for other APIs is calculated from the relative ratio based on sha3_256(b'')

    other_api_call_step_cost =
        API_CALL_STEP_COST * other_api_call_ratio // ScoreApiStepRatio.SHA3_256

    :param context: IconScoreContext instance
    :param ratio: The ratio of a given SCORE API based on ScoreApiStepRatio.SHA3_256
    :return: step_cost for a given SCORE API
    """
    api_call_step: int = context.step_counter.get_step_cost(StepType.API_CALL)
    return api_call_step * ratio // ScoreApiStepRatio.SHA3_256


def revert(message: Optional[str] = None, code: int = 0) -> None:
    """
    Reverts the transaction and breaks.
    All the changes of state DB in current transaction will be rolled back.

    :param message: revert message
    :param code: code
    """
    try:
        if not isinstance(code, int):
            code = int(code)

        if not isinstance(message, str):
            message = str(message)
    except:
        raise InvalidParamsException("Revert error: code or message is invalid")
    else:
        raise IconScoreException(message, code)


def sha3_256(data: bytes) -> bytes:
    """
    Computes sha3_256 hash using the input data

    :param data: input data
    :return: hashed data in bytes
    """
    return _hash("sha3_256", data)


def sha_256(data: bytes) -> bytes:
    """
    Computes sha256 hash using the input data

    :param data: input data
    :return: hashed data in bytes
    """
    return _hash("sha256", data)


def _hash(name: str, data: bytes) -> bytes:
    """Protected hash function

    :param name: hash function name: "sha256" or "sha3_256"
    :param data: data to hash
    :return: hashed data in bytes
    """
    if name not in ("sha3_256", "sha256"):
        raise InvalidParamsException(f"Not supported: {name}")
    if not isinstance(data, bytes):
        raise InvalidParamsException("Invalid dataType")

    context = ContextContainer._get_context()
    assert context

    if context and context.revision >= Revision.THREE.value:
        size = len(data)
        chunks = size // 32
        if size % 32 > 0:
            chunks += 1

        step_cost: int = context.step_counter.get_step_cost(StepType.API_CALL)
        step: int = step_cost + step_cost * chunks // 10

        context.step_counter.consume_step(StepType.API_CALL, step)

    func = getattr(hashlib, name)
    return func(data).digest()


def json_dumps(obj: Any) -> str:
    """
    Converts a python object `obj` to a JSON string

    :param obj: a python object to be converted
    :return: json string
    """
    context = ContextContainer._get_context()
    assert context

    if context and context.revision >= Revision.THREE.value:
        ret: str = json.dumps(obj, separators=(',', ':'))

        step_cost: int = _get_api_call_step_cost(context, ScoreApiStepRatio.JSON_DUMPS)
        step: int = step_cost + step_cost * len(ret.encode(CHARSET_ENCODING)) // 100

        context.step_counter.consume_step(StepType.API_CALL, step)
    else:
        ret: str = json.dumps(obj)

    return ret


def json_loads(src: str) -> Any:
    """
    Parses a JSON string `src` and converts it to a python object

    :param src: a JSON string to be converted
    :return: a python object
    """
    if not isinstance(src, str):
        return None

    context = ContextContainer._get_context()
    assert context

    if context and context.revision >= Revision.THREE.value:
        step_cost: int = _get_api_call_step_cost(context, ScoreApiStepRatio.JSON_LOADS)
        step: int = step_cost + step_cost * len(src.encode(CHARSET_ENCODING)) // 100

        context.step_counter.consume_step(StepType.API_CALL, step)

    return json.loads(src)


def create_address_with_key(public_key: bytes) -> Optional['Address']:
    """Create an address with a given public key

    :param public_key: Public key based on secp256k1
    :return: Address created from a given public key or None if failed
    """
    if not isinstance(public_key, bytes):
        return None

    # 33: prefix(1 byte) + keyBody(32 bytes)
    # 65: prefix(1 byte) + keyBody(64 bytes)
    key_size: int = len(public_key)
    if key_size not in (33, 65):
        return None

    context = ContextContainer._get_context()
    assert context

    if context and context.revision >= Revision.THREE.value:
        if key_size == 33:
            ratio = ScoreApiStepRatio.CREATE_ADDRESS_WITH_COMPRESSED_KEY
        else:
            ratio = ScoreApiStepRatio.CREATE_ADDRESS_WITH_UNCOMPRESSED_KEY

        step: int = _get_api_call_step_cost(context, ratio)
        context.step_counter.consume_step(StepType.API_CALL, step)

    try:
        return _create_address_with_key(public_key)
    except:
        return None


def _create_address_with_key(public_key: bytes) -> Optional['Address']:
    assert isinstance(public_key, bytes)
    assert len(public_key) in (33, 65)

    size = len(public_key)
    prefix: int = public_key[0]

    if size == 33 and prefix in (0x02, 0x03):
        uncompressed_public_key: bytes = _convert_key(public_key, compressed=True)
    elif size == 65 and prefix == 0x04:
        uncompressed_public_key: bytes = public_key
    else:
        return None

    body: bytes = hashlib.sha3_256(uncompressed_public_key[1:]).digest()[-20:]
    return Address(AddressPrefix.EOA, body)


def _convert_key(public_key: bytes, compressed: bool) -> Optional[bytes]:
    """Convert key between compressed and uncompressed keys

    :param public_key: compressed or uncompressed key
    :return: the counterpart key of a given public_key
    """
    public_key_object = PublicKey(public_key)
    return public_key_object.format(compressed=not compressed)


def recover_key(msg_hash: bytes, signature: bytes, compressed: bool = True) -> Optional[bytes]:
    """Returns the public key from message hash and recoverable signature

    :param msg_hash: 32 bytes data
    :param signature: signature_data(64) + recovery_id(1)
    :param compressed: the type of public key to return
    :return: public key recovered from msg_hash and signature
        (compressed: 33 bytes key, uncompressed: 65 bytes key)
    """
    context = ContextContainer._get_context()
    assert context

    if context and context.revision >= Revision.THREE.value:
        step_cost: int = _get_api_call_step_cost(context, ScoreApiStepRatio.RECOVER_KEY)
        context.step_counter.consume_step(StepType.API_CALL, step_cost)

    try:
        return _recover_key(msg_hash, signature, compressed)
    except:
        return None


def _recover_key(msg_hash: bytes, signature: bytes, compressed: bool) -> Optional[bytes]:
    if isinstance(msg_hash, bytes) \
            and len(msg_hash) == 32 \
            and isinstance(signature, bytes) \
            and len(signature) == 65:
        return PublicKey.from_signature_and_message(signature, msg_hash, hasher=None).format(compressed)

    return None


class PRepInfo(object):
    def __init__(self, address: 'Address', delegated: int, name: str):
        self.address = address
        self.delegated = delegated
        self.name = name


def get_main_prep_info() -> Tuple[List[PRepInfo], int]:
    context = ContextContainer._get_context()
    assert context

    term = context.term
    if term is None:
        return [], -1

    prep_info_list: List[PRepInfo] = []
    for prep_snapshot in term.main_preps:
        prep = context.get_prep(prep_snapshot.address)
        prep_info_list.append(PRepInfo(
            prep_snapshot.address,
            prep_snapshot.delegated,
            prep.name
        ))
    return prep_info_list, term.end_block_height


def get_sub_prep_info() -> Tuple[List[PRepInfo], int]:
    context = ContextContainer._get_context()
    assert context

    term = context.term
    if term is None:
        return [], -1

    prep_info_list: List[PRepInfo] = []
    for prep_snapshot in term.sub_preps:
        prep = context.get_prep(prep_snapshot.address)
        prep_info_list.append(PRepInfo(
            prep_snapshot.address,
            prep_snapshot.delegated,
            prep.name
        ))
    return prep_info_list, term.end_block_height


def create_interface_score(addr_to: 'Address',
                           interface_cls: Callable[['Address'], T]) -> T:
        """
        Creates an object, through which you have an access to the designated SCORE’s external functions.

        :param addr_to: SCORE address
        :param interface_cls: interface class
        :return: An instance of given class
        """

        if interface_cls is InterfaceScore:
            raise InvalidInstanceException(FORMAT_IS_NOT_DERIVED_OF_OBJECT.format(InterfaceScore.__name__))
        return interface_cls(addr_to)



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Callable

from iconcommons import Logger
from .data.prep import PRep, PenaltyReason

if TYPE_CHECKING:
    from ..base.address import Address
    from ..iconscore.icon_score_context import IconScoreContext


class PenaltyImposer(object):
    """Check for the block validation statistics of P-Rep and impose a proper penalty on it

    """

    def __init__(self,
                 penalty_grace_period: int,
                 low_productivity_penalty_threshold: int,
                 block_validation_penalty_threshold: int):
        # Low productivity penalty is not imposed during penalty_grace_period
        self._penalty_grace_period: int = penalty_grace_period

        # Unit: percent without fraction
        self._low_productivity_penalty_threshold: int = low_productivity_penalty_threshold

        # Unit: The number of blocks
        self._block_validation_penalty_threshold: int = block_validation_penalty_threshold

        Logger.info(f"[PenaltyImposer Init] "
                    f"penalty_grace_period: {self._penalty_grace_period} "
                    f"low_productivity_penalty_threshold: {self._low_productivity_penalty_threshold} "
                    f"block_validation_penalty_threshold: {self._block_validation_penalty_threshold}")

    def run(self,
            context: 'IconScoreContext',
            prep: 'PRep',
            on_penalty_imposed: Callable[['IconScoreContext', 'Address', 'PenaltyReason'], None]) -> 'PenaltyReason':
        reason: 'PenaltyReason' = PenaltyReason.NONE

        if self._check_block_validation_penalty(prep):
            Logger.info(f"PenaltyImposer statistics({PenaltyReason.BLOCK_VALIDATION}): "
                        f"prep: {prep.address} "
                        f"node_address: {prep.node_address} "
                        f"prep_total_blocks: {prep.total_blocks} "
                        f"prep_block_validation_proportion: {prep.block_validation_proportion}")
            reason = PenaltyReason.BLOCK_VALIDATION
        if self._check_low_productivity_penalty(prep):
            Logger.info(f"PenaltyImposer statistics({PenaltyReason.LOW_PRODUCTIVITY}): "
                        f"prep: {prep.address} "
                        f"node_address: {prep.node_address} "
                        f"prep_total_blocks: {prep.total_blocks} "
                        f"prep_unvalidated_sequence_blocks: {prep.unvalidated_sequence_blocks}")
            reason = PenaltyReason.LOW_PRODUCTIVITY

        if on_penalty_imposed and reason != PenaltyReason.NONE:
            on_penalty_imposed(context, prep.address, reason)

        return reason

    def _check_low_productivity_penalty(self, prep: 'PRep') -> bool:
        return prep.total_blocks > self._penalty_grace_period and \
               prep.block_validation_proportion < self._low_productivity_penalty_threshold

    def _check_block_validation_penalty(self, prep: 'PRep') -> bool:
        return prep.unvalidated_sequence_blocks >= self._block_validation_penalty_threshold



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import decimal
from typing import TYPE_CHECKING

from .engine import Engine as IISSEngine
from .engine import Method as IISSMethod
from .storage import Storage as IISSStorage
from ..icon_constant import Revision

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext
    from ..prep.data.prep import PRep


def check_decentralization_condition(context: 'IconScoreContext') -> bool:
    if context.revision < Revision.DECENTRALIZATION.value or context.is_decentralized():
        # If revision is less than REV_DECENTRALIZATION or
        # network has been already decentralized
        return False

    context.update_dirty_prep_batch()

    """ICON network decentralize when the last prep of main prep count ( default: 22th )
    get delegation more than some value( default: total-supply * 0.002icx )"""
    if context.preps.size(active_prep_only=True) >= context.main_prep_count:
        minimum_delegate = get_minimum_delegate_for_bottom_prep(context)
        bottom_prep: 'PRep' = context.preps.get_by_index(context.main_prep_count - 1)
        return bottom_prep.delegated >= minimum_delegate
    return False


def get_minimum_delegate_for_bottom_prep(context: 'IconScoreContext') -> int:
    """Minimum delegate default value = total_supply * 0.002 ICX"""
    assert 1.0 > context.decentralize_trigger >= 0
    numerator, denominator = _split_float_to_numerator_and_denominator(context.decentralize_trigger)
    if numerator == 0:
        return 0

    total_supply: int = context.storage.icx.get_total_supply(context)
    minimum_delegate: int = total_supply * numerator // denominator
    return minimum_delegate


def _split_float_to_numerator_and_denominator(float_data: float) -> tuple:
    assert 1.0 > float_data >= 0
    str_float: str = format(decimal.Decimal(str(float_data)), 'f')
    str_decimal: str = str_float[str_float.find('.') + 1:]
    numerator = int(str_decimal)
    denominator = 10 ** len(str_decimal)
    return numerator, denominator



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import TYPE_CHECKING

from iconcommons import Logger

from .icon_score_deployer import IconScoreDeployer
from .utils import remove_path, get_score_path
from ..base.ComponentBase import EngineBase
from ..base.address import (
    Address, SYSTEM_SCORE_ADDRESS,
    generate_score_address_for_tbears, generate_score_address,
)
from ..base.exception import AccessDeniedException, InvalidParamsException
from ..base.message import Message
from ..base.type_converter import TypeConverter
from ..icon_constant import DeployType
from ..icon_constant import IconServiceFlag, ICON_DEPLOY_LOG_TAG, Revision
from ..iconscore.icon_score_api_generator import ScoreApiGenerator
from ..iconscore.icon_score_context_util import IconScoreContextUtil
from ..iconscore.icon_score_mapper_object import IconScoreInfo
from ..iconscore.icon_score_step import StepType, get_deploy_content_size
from ..iconscore.typing.conversion import convert_score_parameters
from ..iconscore.typing.element import check_score_flag
from ..iconscore.typing.element import normalize_signature
from ..iconscore.utils import get_score_deploy_path
from ..utils import is_builtin_score

if TYPE_CHECKING:
    from .storage import IconScoreDeployInfo
    from .storage import IconScoreDeployTXParams
    from ..iconscore.icon_score_context import IconScoreContext
    from ..iconscore.icon_score_mapper import IconScoreMapper
    from ..iconscore.icon_score_base import IconScoreBase


class Engine(EngineBase):
    """It handles transactions to install, update and audit a SCORE
    """

    def invoke(self, context: 'IconScoreContext', to: 'Address', data: dict) -> 'Address':
        """Handle data contained in icx_sendTransaction message
        :param context:
        :param to: If 'to' is SYSTEM_SCORE_ADDRESS, install SCORE. Otherwise update SCORE
        :param data: SCORE deploy data
        :return SCORE address
        """
        if to == SYSTEM_SCORE_ADDRESS:
            # SCORE install
            content_type = data.get("contentType")
            if content_type == "application/tbears":
                path: str = data.get("content")
                score_address = generate_score_address_for_tbears(path)
            else:
                score_address = generate_score_address(
                    context.tx.origin, context.tx.timestamp, context.tx.nonce
                )
            deploy_info = IconScoreContextUtil.get_deploy_info(context, score_address)
            if deploy_info is not None:
                raise AccessDeniedException(
                    f"SCORE address already in use: {score_address}"
                )

            context.step_counter.apply_step(StepType.CONTRACT_CREATE, 1)
        else:
            # SCORE update
            score_address = to
            context.step_counter.apply_step(StepType.CONTRACT_UPDATE, 1)

        content_size = get_deploy_content_size(
            context.revision, data.get("content", None)
        )
        context.step_counter.apply_step(StepType.CONTRACT_SET, content_size)

        self._invoke(
            context=context, to=to, icon_score_address=score_address, data=data
        )

        return score_address

    def _invoke(self, context: "IconScoreContext", to: "Address",
                icon_score_address: "Address", data: dict) -> None:
        assert icon_score_address is not None
        assert icon_score_address != SYSTEM_SCORE_ADDRESS
        assert icon_score_address.is_contract

        if icon_score_address in (None, SYSTEM_SCORE_ADDRESS):
            raise InvalidParamsException(f'Invalid SCORE address: {icon_score_address}')

        try:
            deploy_type: 'DeployType' = \
                DeployType.INSTALL if to == SYSTEM_SCORE_ADDRESS else DeployType.UPDATE

            context.storage.deploy.put_deploy_info_and_tx_params(
                context, icon_score_address, deploy_type,
                context.tx.origin, context.tx.hash, data)

            if not self._is_audit_needed(context, icon_score_address):
                self.deploy(context, context.tx.hash)

        except BaseException as e:
            Logger.warning('Failed to write deploy info and tx params', ICON_DEPLOY_LOG_TAG)
            raise e

    @staticmethod
    def _is_audit_needed(context: 'IconScoreContext', score_address: Address) -> bool:
        """Check whether audit process is needed or not
        :param context:
        :param score_address:
        :return: True(needed) False(not needed)
        """
        if context.revision >= Revision.TWO.value:
            is_system_score = is_builtin_score(str(score_address))
        else:
            is_system_score = False

        # FiXME: SCORE owner check should be done before calling self._is_audit_needed().
        is_owner: bool = context.tx.origin == IconScoreContextUtil.get_owner(context, score_address)
        is_audit_enabled: bool = IconScoreContextUtil.is_service_flag_on(context, IconServiceFlag.AUDIT)

        return is_audit_enabled and not (is_system_score and is_owner)

    def deploy(self, context: 'IconScoreContext', tx_hash: bytes) -> None:
        """
        1. Convert a content from hex string to bytes
        2. Decompress zipped SCORE code and write it to filesystem
        3. Import the decompressed SCORE code
        4. Create a SCORE instance from the code
        5. Run on_install() or on_update() method in the SCORE
        6. Update the deployed SCORE info to stateDB
        :param context:
        :param tx_hash:
        """

        tx_params: 'IconScoreDeployTXParams' =\
            context.storage.deploy.get_deploy_tx_params(context, tx_hash)
        if tx_params is None:
            raise InvalidParamsException(f'tx_params is None: 0x{tx_hash.hex()}')

        score_address: 'Address' = tx_params.score_address
        tmp_current = context.current_address
        context.current_address = score_address
        self._score_deploy(context, tx_params)
        context.storage.deploy.update_score_info(context, score_address, tx_hash)
        context.current_address = tmp_current

    def _score_deploy(self, context: 'IconScoreContext', tx_params: 'IconScoreDeployTXParams'):
        """
        :param tx_params: use deploy_data from IconScoreDeployTxParams info
        :return:
        """

        data: dict = tx_params.deploy_data
        content_type: str = data.get('contentType')

        if content_type == 'application/tbears':
            if not context.legacy_tbears_mode:
                raise InvalidParamsException(f'Invalid contentType: application/tbears')
        elif content_type == 'application/zip':
            data['content'] = bytes.fromhex(data['content'][2:])
        else:
            raise InvalidParamsException(
                f'Invalid contentType: {content_type}')

        self._on_deploy(context, tx_params)

    def _on_deploy(self,
                   context: 'IconScoreContext',
                   tx_params: 'IconScoreDeployTXParams') -> None:
        """
        Decompress a SCORE zip file and write them to file system
        Create a SCORE instance from SCORE class
        Call a SCORE initialization function (on_install or on_update)
        :param tx_params: use deploy_data, score_address, tx_hash, deploy_type from IconScoreDeployTxParams
        :return:
        """

        data = tx_params.deploy_data
        score_address = tx_params.score_address
        params: dict = data.get('params', {})

        deploy_info: 'IconScoreDeployInfo' = context.storage.deploy.get_deploy_info(context, tx_params.score_address)
        next_tx_hash: bytes = deploy_info.next_tx_hash

        self._write_score_to_filesystem(context, score_address, next_tx_hash, data)

        backup_msg = context.msg
        backup_tx = context.tx
        new_tx_score_mapper: dict = {}

        try:
            IconScoreContextUtil.validate_score_package(context, score_address, next_tx_hash)

            score_info: 'IconScoreInfo' =\
                self._create_score_info(context, score_address, next_tx_hash)

            if context.revision >= Revision.STRICT_SCORE_DECORATOR_CHECK.value:
                check_score_flag(score_info.score_class)

            # score_info.get_score() returns a cached or created score instance
            # according to context.revision.
            score: 'IconScoreBase' = score_info.get_score(context.revision)

            # check_on_deploy will be done by normalize_signature()
            # since Revision.THREE
            if context.revision < Revision.THREE.value:
                ScoreApiGenerator.check_on_deploy(context, score)

            # owner is set in IconScoreBase.__init__()
            context.msg = Message(sender=score.owner)
            context.tx = None

            self._initialize_score(context, tx_params.deploy_type, score, params)
            new_tx_score_mapper[score_address] = score_info
        except BaseException as e:
            Logger.warning(f'Failed to deploy a SCORE: {score_address}', ICON_DEPLOY_LOG_TAG)
            raise e
        finally:
            context.msg = backup_msg
            context.tx = backup_tx
            self._update_new_score_mapper(context.new_icon_score_mapper, new_tx_score_mapper)

    @classmethod
    def _update_new_score_mapper(cls, block_mapper: 'IconScoreMapper', tx_mapper: dict):
        for address, score_info in tx_mapper.items():
            block_mapper[address] = score_info

    def _write_score_to_filesystem(self, context: 'IconScoreContext',
                                   score_address: 'Address', tx_hash: bytes, deploy_data: dict):

        content_type: str = deploy_data.get('contentType')
        content = deploy_data.get('content')

        if content_type == 'application/tbears':
            write_score_to_score_deploy_path: callable =\
                self._write_score_to_score_deploy_path_on_tbears_mode
        else:
            write_score_to_score_deploy_path: callable =\
                self._write_score_to_score_deploy_path

        write_score_to_score_deploy_path(context, score_address, tx_hash, content)

    @staticmethod
    def _create_score_info(context: 'IconScoreContext',
                           score_address: 'Address', tx_hash: bytes) -> 'IconScoreInfo':
        """Create the score_info instance associated with the SCORE to deploy
        :param context:
        :param score_address:
        :param tx_hash:
        :return:
        """
        current_score_info: 'IconScoreInfo' = IconScoreContextUtil.get_score_info(context, score_address)

        # Reuse score_db if it has already existed.
        score_db = None
        if current_score_info is not None:
            score_db = current_score_info.score_db

        return IconScoreContextUtil.create_score_info(context, score_address, tx_hash, score_db)

    @staticmethod
    def _write_score_to_score_deploy_path_on_tbears_mode(
            context: 'IconScoreContext', score_address: 'Address', tx_hash: bytes, content: bytes):
        score_root_path: str = context.score_root_path
        score_path: str = get_score_path(score_root_path, score_address)
        os.makedirs(score_path, exist_ok=True)

        score_deploy_path: str = get_score_deploy_path(score_root_path, score_address, tx_hash)
        try:
            os.symlink(content, score_deploy_path, target_is_directory=True)
        except FileExistsError:
            pass

    @staticmethod
    def _write_score_to_score_deploy_path(context: 'IconScoreContext',
                                          score_address: 'Address', tx_hash: bytes, content: bytes):
        """Write SCORE code to file system

        :param context: IconScoreContext instance
        :param score_address: score address
        :param tx_hash: transaction hash
        :param content: zipped SCORE code data
        :return:
        """
        revision: int = context.revision

        # score_root_path is the directory which contains all deployed scores.
        score_root_path: str = context.score_root_path
        score_deploy_path: str = get_score_deploy_path(score_root_path, score_address, tx_hash)

        if revision >= Revision.THREE.value:
            # If the path to deploy a score has been present, remove it before deploying.
            score_root_path: str = context.score_root_path
            score_path: str =\
                os.path.join(score_root_path, score_address.to_bytes().hex(), f'0x{tx_hash.hex()}')
            remove_path(score_path)

        if revision >= Revision.TWO.value:
            IconScoreDeployer.deploy(score_deploy_path, content, revision)
        else:
            IconScoreDeployer.deploy_legacy(score_deploy_path, content)

    @staticmethod
    def _initialize_score(
            context: 'IconScoreContext',
            deploy_type: DeployType,
            score: 'IconScoreBase', params: dict):
        """Call on_install() or on_update() of a SCORE
        only once when installing or updating it
        :param deploy_type: DeployType.INSTALL or DeployType.UPDATE
        :param score: SCORE to install or update
        :param params: parameters passed to on_install or on_update()
        """
        if deploy_type == DeployType.INSTALL:
            on_init = score.on_install
        elif deploy_type == DeployType.UPDATE:
            on_init = score.on_update
        else:
            raise InvalidParamsException(f'Invalid deployType: {deploy_type}')

        if context.revision < Revision.THREE.value:
            TypeConverter.adjust_params_to_method(on_init, params)
        else:
            signatures = [None, None]
            # Do not allow **kwargs, *args used in on_install() and on_update() of score
            signatures[DeployType.INSTALL.value] = normalize_signature(score.on_install)
            signatures[DeployType.UPDATE.value] = normalize_signature(score.on_update)

            params = convert_score_parameters(params, signatures[deploy_type.value])

        on_init(**params)



# -*- coding: utf-8 -*-
# Copyright 2020 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from inspect import (
    Signature,
    Parameter,
)
from typing import Optional, Tuple, Dict, Any, List, Mapping, Union

from . import (
    get_origin,
    get_args,
    get_annotations,
    is_base_type,
    is_struct,
    isinstance_ex,
)
from ...base.exception import InvalidParamsException


def verify_internal_call_arguments(sig: Signature, args: Optional[Tuple], kwargs: Optional[Dict]):
    """Check if argument types matches to type hints in sig.parameters

    :param sig: normalized signature of method in SCORE
    :param args: args in tuple
    :param kwargs:
    :return:
    """
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}

    params: Dict[str, Any] = {}
    parameters = sig.parameters

    merge_arguments(params, parameters, args, kwargs)
    set_default_value_to_params(params, parameters)

    for name, parameter in parameters.items():
        if name not in params:
            raise InvalidParamsException(f"Argument not found: name={name}")

        type_hint = parameter.annotation
        value = params[name]

        try:
            verify_type_hint(value, type_hint)
        except TypeError:
            raise InvalidParamsException(
                f"Type mismatch: name={name} type_hint={type_hint} value_type={type(value)}")


def merge_arguments(
        params: Dict[str, Any],
        parameters: Mapping[str, Parameter],
        args: Optional[Tuple], kwargs: Optional[Dict]):
    """Merge args and kwargs to a dictionary

    Type checking and parameter default value will be handled in the next phase

    :param params: dictionary which will contain merged arguments from args and kwargs
    :param parameters: function signature
    :param args: arguments in tuple
    :param kwargs: keyword arguments in dict
    :return:
    """
    if len(args) > len(parameters):
        raise InvalidParamsException(f"Too many arguments")

    for arg, k in zip(args, parameters):
        params[k] = arg

    for k in kwargs:
        if k in params:
            raise InvalidParamsException(f"Duplicated argument: name={k} value={kwargs[k]}")

        if k not in parameters:
            raise InvalidParamsException(f"Invalid argument: name={k} value={kwargs[k]}")

        params[k] = kwargs[k]


def set_default_value_to_params(params: Dict[str, Any], parameters: Mapping[str, Parameter]):
    """Set default parameter value to missing parameter

    If No default parameter value is available for missing argument, an exception is raised

    :param params:
    :param parameters:
    :return:
    """
    if len(params) == len(parameters):
        return

    # Set default parameter values to missing arguments
    for k in parameters:
        if k in params:
            continue

        parameter = parameters[k]
        if parameter.default is Parameter.empty:
            raise InvalidParamsException(
                f"Missing argument: name={k} type={parameter.annotation}")

        params[k] = parameter.default


def verify_type_hint(value: Any, type_hint: type):
    origin: type = get_origin(type_hint)

    if is_base_type(origin):
        if not isinstance_ex(value, origin):
            raise TypeError
    elif is_struct(origin):
        verify_struct_type_hint(value, type_hint)
    elif origin is list:
        verify_list_type_hint(value, type_hint)
    elif origin is dict:
        verify_dict_type_hint(value, type_hint)
    elif origin is Union:
        verify_union_type_hint(value, type_hint)
    else:
        raise TypeError


def verify_struct_type_hint(value: Dict[str, Any], type_hint: type):
    annotations = get_annotations(type_hint, None)
    assert annotations is not None

    for name, type_hint in annotations.items():
        if name not in value:
            raise InvalidParamsException(f"Missing field in struct: name={name}")

        verify_type_hint(value[name], type_hint)


def verify_list_type_hint(values: List[Any], type_hint: type):
    args = get_args(type_hint)

    for value in values:
        verify_type_hint(value, args[0])


def verify_dict_type_hint(values: Dict[str, Any], type_hint: type):
    args = get_args(type_hint)
    key_type_hint = args[0]
    value_type_hint = args[1]

    assert key_type_hint is str

    for k, v in values.items():
        if not isinstance_ex(k, key_type_hint):
            raise TypeError

        verify_type_hint(v, value_type_hint)


def verify_union_type_hint(value: Union[Any, None], type_hint: type):
    if value is None:
        return

    args = get_args(type_hint)
    verify_type_hint(value, args[0])



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from enum import IntEnum
from typing import Any, TYPE_CHECKING, List, Optional

from ...base.exception import InvalidParamsException
from ...icon_constant import DATA_BYTE_ORDER, RC_DB_VERSION_2, RC_DB_VERSION_0
from ...utils.msgpack_for_ipc import MsgPackForIpc, TypeTag

if TYPE_CHECKING:
    from ...base.address import Address


class TxType(IntEnum):
    DELEGATION = 0
    PREP_REGISTER = 1
    PREP_UNREGISTER = 2
    INVALID = 99


class Data:
    @abstractmethod
    def make_key(self, *args, **kwargs) -> bytes:
        pass

    @abstractmethod
    def make_value(self) -> bytes:
        pass

    @staticmethod
    def from_bytes(*args, **kwargs) -> 'Data':
        pass


class Header(Data):
    PREFIX = b'HD'

    def __init__(self):
        self.version: int = 0
        self.block_height: int = 0
        self.revision: int = 0

    def make_key(self) -> bytes:
        return self.PREFIX

    def make_value(self) -> bytes:
        data = [
            self.version,
            self.block_height
        ]
        if self.version >= RC_DB_VERSION_2:
            # Added value in version 2
            data.append(self.revision)

        return MsgPackForIpc.dumps(data)

    @staticmethod
    def from_bytes(value: bytes) -> 'Header':
        data_list: list = MsgPackForIpc.loads(value)
        version: int = data_list[0]
        if version == RC_DB_VERSION_2:
            return Header._from_bytes_v2(data_list)
        elif version == RC_DB_VERSION_0:
            return Header._from_bytes_v1(data_list)

    @staticmethod
    def _from_bytes_v1(data_list: list) -> 'Header':
        obj = Header()
        obj.version: int = data_list[0]
        obj.block_height: int = data_list[1]
        return obj

    @staticmethod
    def _from_bytes_v2(data_list: list) -> 'Header':
        obj = Header()
        obj.version: int = data_list[0]
        obj.block_height: int = data_list[1]
        obj.revision: int = data_list[2]
        return obj

    def __str__(self):
        info: str = f"[{self.PREFIX}] version: {self.version}, block_height: {self.block_height} "

        if self.version >= RC_DB_VERSION_2:
            info += f"revision: {self.revision} "
        return info


class GovernanceVariable(Data):
    PREFIX = b'GV'

    def __init__(self):
        # key
        self.block_height: int = 0

        # value
        self.version: int = 0
        self.calculated_irep: int = 0
        self.reward_rep: int = 0
        self.config_main_prep_count: int = 0
        self.config_sub_prep_count: int = 0

    def make_key(self) -> bytes:
        block_height: bytes = self.block_height.to_bytes(8, byteorder=DATA_BYTE_ORDER)
        return self.PREFIX + block_height

    def make_value(self) -> bytes:
        data = [
            self.calculated_irep,
            self.reward_rep,
        ]
        if self.version >= RC_DB_VERSION_2:
            # Added value in version 2
            data.append(self.config_main_prep_count)
            data.append(self.config_sub_prep_count)

        return MsgPackForIpc.dumps(data)

    @staticmethod
    def from_bytes(key: bytes, value: bytes) -> 'GovernanceVariable':
        # Method for debugging
        data_list: list = MsgPackForIpc.loads(value)

        # need to be refactor
        if len(data_list) > 2:
            return GovernanceVariable._from_bytes_v2(key, data_list)
        else:
            return GovernanceVariable._from_bytes_v1(key, data_list)

    @staticmethod
    def _from_bytes_v1(key: bytes, data_list: list) -> 'GovernanceVariable':
        obj = GovernanceVariable()
        obj.block_height: int = int.from_bytes(key[2:], DATA_BYTE_ORDER)
        obj.version: int = RC_DB_VERSION_0
        obj.calculated_irep: int = data_list[0]
        obj.reward_rep: int = data_list[1]
        return obj

    @staticmethod
    def _from_bytes_v2(key: bytes, data_list: list) -> 'GovernanceVariable':
        obj = GovernanceVariable()
        obj.block_height: int = int.from_bytes(key[2:], DATA_BYTE_ORDER)
        obj.version: int = RC_DB_VERSION_2
        obj.calculated_irep: int = data_list[0]
        obj.reward_rep: int = data_list[1]
        obj.config_main_prep_count: int = data_list[2]
        obj.config_sub_prep_count: int = data_list[3]
        return obj

    def __str__(self):
        info: str = f"[{self.PREFIX}] key: {self.block_height}," \
                    f" calculated_irep: {self.calculated_irep}, reward_rep: {self.reward_rep}"
        if self.version >= RC_DB_VERSION_2:
            info += f"config_main_prep_count: {self.config_main_prep_count}, " \
                    f"config_sub_prep_count: {self.config_sub_prep_count}"
        return info


def make_block_produce_info_key(block_height: int) -> bytes:
    return BlockProduceInfoData.PREFIX + block_height.to_bytes(8, byteorder=DATA_BYTE_ORDER)


class BlockProduceInfoData(Data):
    PREFIX = b'BP'

    def __init__(self):
        # key
        self.block_height: int = 0

        # value
        self.block_generator: Optional['Address'] = None
        self.block_validator_list: Optional[List['Address']] = None

    def make_key(self) -> bytes:
        return make_block_produce_info_key(self.block_height)

    def make_value(self) -> bytes:
        data = [
            MsgPackForIpc.encode(self.block_generator),
            [MsgPackForIpc.encode(validator_address) for validator_address in self.block_validator_list]
        ]
        return MsgPackForIpc.dumps(data)

    @staticmethod
    def from_bytes(key: bytes, value: bytes) -> 'BlockProduceInfoData':
        # Method for debugging
        data_list: list = MsgPackForIpc.loads(value)
        obj = BlockProduceInfoData()
        obj.block_height: int = int.from_bytes(key[2:], DATA_BYTE_ORDER)
        obj.block_generator: 'Address' = MsgPackForIpc.decode(TypeTag.ADDRESS, data_list[0])

        obj.block_validator_list: list = [MsgPackForIpc.decode(TypeTag.ADDRESS, bytes_address)
                                          for bytes_address in data_list[1]]
        return obj

    def __str__(self):
        return f"[{self.PREFIX}] " \
               f"key: {self.block_height}, " \
               f"block_generator: {str(self.block_generator)}, " \
               f"block_validators: {[str(addr) for addr in self.block_validator_list]}"


class PRepsData(Data):
    PREFIX = b'PR'

    def __init__(self):
        # key
        self.block_height: int = 0

        # value
        self.total_delegation: int = 0
        self.prep_list: Optional[List['DelegationInfo']] = None

    def make_key(self) -> bytes:
        block_height: bytes = self.block_height.to_bytes(8, byteorder=DATA_BYTE_ORDER)
        return self.PREFIX + block_height

    def make_value(self) -> bytes:
        encoded_prep_list = [[MsgPackForIpc.encode(delegation_info.address),
                              MsgPackForIpc.encode(delegation_info.value)] for delegation_info in self.prep_list]
        data = [
            MsgPackForIpc.encode(self.total_delegation),
            encoded_prep_list
        ]
        return MsgPackForIpc.dumps(data)

    @staticmethod
    def from_bytes(key: bytes, value: bytes) -> 'PRepsData':
        # Method for debugging
        data_list: list = MsgPackForIpc.loads(value)
        obj = PRepsData()
        obj.prep_list = []
        obj.block_height: int = int.from_bytes(key[2:], DATA_BYTE_ORDER)
        obj.total_delegation = MsgPackForIpc.decode(TypeTag.INT, data_list[0])
        prep_list: list = [
            [MsgPackForIpc.decode(TypeTag.ADDRESS, delegation_info[0]),
             MsgPackForIpc.decode(TypeTag.INT, delegation_info[1])]
            for delegation_info in data_list[1]
        ]
        for prep in prep_list:
            del_info = DelegationInfo()
            del_info.address = prep[0]
            del_info.value = prep[1]
            obj.prep_list.append(del_info)
        return obj

    def __str__(self):
        return f"[{self.PREFIX}] " \
               f"key: {self.block_height}, total_delegation: {str(self.total_delegation)}"


class TxData(Data):
    PREFIX = b'TX'

    def __init__(self):
        self.address: 'Address' = None
        self.block_height: int = 0
        self.type: 'TxType' = TxType.INVALID
        self.data: 'Tx' = None

    def make_key(self, index: int) -> bytes:
        tx_index: bytes = index.to_bytes(8, byteorder=DATA_BYTE_ORDER)
        return self.PREFIX + tx_index

    def make_value(self) -> bytes:
        tx_type: 'TxType' = self.type
        tx_data: 'Tx' = self.data

        if isinstance(tx_data, Tx):
            tx_data_type = tx_data.get_type()
            if tx_type == TxType.INVALID:
                tx_type = tx_data_type
            elif tx_type != tx_data_type:
                raise InvalidParamsException(f"Mismatch TxType: {tx_type}")
        else:
            raise InvalidParamsException(f"Invalid TxData: {tx_data}")

        data = [
            MsgPackForIpc.encode(self.address),
            self.block_height,
            tx_type,
            tx_data.encode()
        ]

        return MsgPackForIpc.dumps(data)

    @staticmethod
    def from_bytes(value: bytes) -> 'TxData':
        # Method for debugging
        data_list: list = MsgPackForIpc.loads(value)
        obj = TxData()
        obj.address: 'Address' = MsgPackForIpc.decode(TypeTag.ADDRESS, data_list[0])
        obj.block_height: int = data_list[1]
        obj.type: 'TxType' = TxType(data_list[2])
        obj.data: 'Tx' = TxData._covert_tx_data(obj.type, data_list[3])
        return obj

    @staticmethod
    def _covert_tx_data(tx_type: 'TxType', data: tuple) -> Any:
        if tx_type == TxType.DELEGATION:
            return DelegationTx.decode(data)
        elif tx_type == TxType.PREP_REGISTER:
            return PRepRegisterTx.decode(data)
        elif tx_type == TxType.PREP_UNREGISTER:
            return PRepUnregisterTx.decode(data)
        else:
            raise InvalidParamsException(f"InvalidParams TxType: {tx_type}")


class Tx(object, metaclass=ABCMeta):
    @abstractmethod
    def get_type(self) -> 'TxType':
        pass

    @abstractmethod
    def encode(self) -> list:
        pass

    @staticmethod
    @abstractmethod
    def decode(data: list) -> Any:
        pass


class DelegationTx(Tx):
    def __init__(self):
        self.delegation_info: List['DelegationInfo'] = []

    def get_type(self) -> 'TxType':
        return TxType.DELEGATION

    def encode(self) -> tuple:
        data = [x.encode() for x in self.delegation_info]
        return MsgPackForIpc.encode_any(data)

    @staticmethod
    def decode(data: tuple) -> 'DelegationTx':
        data_list: list = MsgPackForIpc.decode_any(data)
        obj = DelegationTx()
        obj.delegation_info: list = [DelegationInfo.decode(x) for x in data_list]
        return obj


class DelegationInfo(object):
    def __init__(self):
        self.address: 'Address' = None
        self.value: int = None

    def encode(self) -> list:
        return [self.address, self.value]

    @staticmethod
    def decode(data: list) -> 'DelegationInfo':
        obj = DelegationInfo()
        obj.address: 'Address' = data[0]
        obj.value: int = data[1]
        return obj


class PRepRegisterTx(Tx):
    def __init__(self):
        pass

    def get_type(self) -> 'TxType':
        return TxType.PREP_REGISTER

    def encode(self) -> tuple:
        return MsgPackForIpc.encode_any(None)

    @staticmethod
    def decode(data: tuple) -> 'PRepRegisterTx':
        obj = PRepRegisterTx()
        return obj


class PRepUnregisterTx(Tx):
    def __init__(self):
        pass

    def get_type(self) -> 'TxType':
        return TxType.PREP_UNREGISTER

    def encode(self) -> tuple:
        return MsgPackForIpc.encode_any(None)

    @staticmethod
    def decode(data: tuple) -> 'PRepUnregisterTx':
        obj = PRepUnregisterTx()
        return obj



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from ..base.address import Address
from ..utils.msgpack_for_db import MsgPackForDB


class Deposit(object):
    """
    Deposit Information Class
    implementing functions to serialize, deserialize and convert to dict type.
    """
    _VERSION = 0

    # Percentage of the minimum remaining deposit amount to pay fee.
    # The minimum remaining amount of a single deposit is 10 percent of amount of the deposit
    _MIN_REMAINING_PROPORTION = 10

    _EXPOSING_ITEM_KEYS = (
        'id',
        'sender',
        'deposit_amount',
        'deposit_used',
        'created',
        'expires',
        'virtual_step_issued',
        'virtual_step_used'
    )

    def __init__(self, deposit_id: bytes = None, score_address: 'Address' = None, sender: 'Address' = None,
                 deposit_amount: int = 0, deposit_used: int = 0, created: int = 0, expires: int = -1,
                 virtual_step_issued: int = 0, virtual_step_used: int = 0,
                 prev_id: bytes = None, next_id: bytes = None):
        self.version = self._VERSION
        # deposit id, should be tx hash of deposit transaction
        self.id = deposit_id
        # target SCORE address
        self.score_address = score_address
        # sender address
        self.sender = sender
        # deposit amount of ICXs in loop
        self.deposit_amount = deposit_amount
        # used amount of deposited ICXs in loop
        self.deposit_used = deposit_used
        # created time in block
        self.created = created
        # expires time in block
        self.expires = expires
        # issued amount of virtual STEPs
        self.virtual_step_issued = virtual_step_issued
        # used amount of virtual STEPs
        self.virtual_step_used = virtual_step_used
        # previous id of this deposit
        self.prev_id = prev_id
        # next id of this deposit
        self.next_id = next_id

    @staticmethod
    def from_bytes(buf: bytes):
        """Creates Deposit object from bytes data.

        :param buf: deposit info in bytes
        :return: deposit object
        """
        data: list = MsgPackForDB.loads(buf)

        deposit = Deposit()
        deposit.version = data[0]
        deposit.score_address = data[1]
        deposit.sender = data[2]
        deposit.deposit_amount = data[3]
        deposit.deposit_used = data[4]
        deposit.created = data[5]
        deposit.expires = data[6]
        deposit.virtual_step_issued = data[7]
        deposit.virtual_step_used = data[8]
        deposit.prev_id = data[9]
        deposit.next_id = data[10]

        return deposit

    def to_bytes(self) -> bytes:
        """Converts Deposit object into bytes.

        :return: deposit info in bytes
        """
        data: list = [self.version,
                      self.score_address,
                      self.sender,
                      self.deposit_amount,
                      self.deposit_used,
                      self.created,
                      self.expires,
                      self.virtual_step_issued,
                      self.virtual_step_used,
                      self.prev_id,
                      self.next_id]

        return MsgPackForDB.dumps(data)

    def to_dict(self, casing: Optional = None) -> dict:
        """Returns properties as dict.

        :param casing: a kind of functions to convert one casing notation to another
        :return: deposit info in dict
        """
        new_dict = {}
        for key in self._EXPOSING_ITEM_KEYS:
            value = self.__dict__[key]
            if value is not None:
                new_key = casing(key) if casing else key
                new_dict[new_key] = value

        return new_dict

    def __eq__(self, other) -> bool:
        """operator == overriding

        :param other: (Deposit)
        """
        return isinstance(other, Deposit) \
            and self.version == other.version \
            and self.score_address == other.score_address \
            and self.sender == other.sender \
            and self.deposit_amount == other.deposit_amount \
            and self.deposit_used == other.deposit_used \
            and self.created == other.created \
            and self.expires == other.expires \
            and self.virtual_step_issued == other.virtual_step_issued \
            and self.virtual_step_used == other.virtual_step_used \
            and self.prev_id == other.prev_id \
            and self.next_id == other.next_id

    def __ne__(self, other) -> bool:
        """operator != overriding

        :param other: (Deposit)
        """
        return not self.__eq__(other)

    def consume_virtual_step(self, step: int):
        """
        Consume the given step from the virtual step
        """
        self.virtual_step_used += step

    def consume_deposit(self, icx: int):
        """
        Consume the given icx from the deposit
        """
        self.deposit_used += icx

    @property
    def remaining_virtual_step(self):
        """
        the amount of available virtual step
        """
        return self.virtual_step_issued - self.virtual_step_used

    @property
    def remaining_deposit(self):
        """
        the amount of available deposit for fees
        """
        return self.deposit_amount - self.deposit_used

    @property
    def min_remaining_deposit(self):
        """
        the minimum remaining deposit amount
        """
        return self.deposit_amount * self._MIN_REMAINING_PROPORTION // 100



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from enum import IntEnum

import msgpack

from iconservice.base.address import Address
from iconservice.utils import bytes_to_hex
from iconservice.utils.msgpack_for_ipc import MsgPackForIpc, TypeTag

_next_msg_id: int = 1


def reset_next_msg_id(msg_id: int):
    """Only used for unittest

    :param msg_id:
    :return:
    """
    global _next_msg_id
    _next_msg_id = msg_id


def _get_next_msg_id() -> int:
    global _next_msg_id

    msg_id: int = _next_msg_id
    _next_msg_id = max(1, (msg_id + 1) % 0xffffffff)

    return msg_id


class MessageType(IntEnum):
    NONE = -1
    VERSION = 0
    CLAIM = 1
    QUERY = 2
    CALCULATE = 3
    COMMIT_BLOCK = 4
    COMMIT_CLAIM = 5
    QUERY_CALCULATE_STATUS = 6
    QUERY_CALCULATE_RESULT = 7
    ROLLBACK = 8
    INIT = 9
    READY = 100
    CALCULATE_DONE = 101


class Request(metaclass=ABCMeta):
    def __init__(self, msg_type: 'MessageType'):
        self.msg_type = msg_type
        self.msg_id = _get_next_msg_id()

    @abstractmethod
    def _to_list(self) -> tuple:
        return None,

    def to_bytes(self) -> bytes:
        items: tuple = self._to_list()
        return msgpack.dumps(items)


class Response(metaclass=ABCMeta):
    def __init__(self):
        self.msg_id = -1

    @staticmethod
    @abstractmethod
    def from_list(items: list) -> 'Response':
        pass

    def is_notification(self):
        return self.msg_id == 0


class VersionRequest(Request):
    def __init__(self):
        super().__init__(MessageType.VERSION)

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id

    def __str__(self) -> str:
        return f"{self.msg_type.name}({self.msg_id})"


class VersionResponse(Response):
    MSG_TYPE = MessageType.VERSION

    def __init__(self, msg_id: int, version: int, block_height: int):
        super().__init__()

        self.msg_id: int = msg_id
        self.version: int = version
        self.block_height: int = block_height

    def __str__(self):
        return f"VERSION({self.msg_id}, {self.version}, {self.block_height})"

    @staticmethod
    def from_list(items: list) -> 'VersionResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        version: int = payload[0]
        block_height: int = payload[1]

        return VersionResponse(msg_id, version, block_height)


class ClaimRequest(Request):
    def __init__(self, address: 'Address', block_height: int, block_hash: bytes, tx_index: int, tx_hash: bytes):
        super().__init__(MessageType.CLAIM)

        self.address = address
        self.block_height = block_height
        self.block_hash = block_hash
        self.tx_index = tx_index
        self.tx_hash = tx_hash

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id,\
               (
                   self.address.to_bytes_including_prefix(),
                   self.block_height, self.block_hash,
                   self.tx_index, self.tx_hash
               )

    def __str__(self) -> str:
        return \
            f"{self.msg_type.name}({self.msg_id}, " \
            f"{self.address}, " \
            f"{self.block_height}, {bytes_to_hex(self.block_hash)}), " \
            f"{self.tx_index}, {bytes_to_hex(self.tx_hash)}"


class ClaimResponse(Response):
    MSG_TYPE = MessageType.CLAIM

    def __init__(self, msg_id: int, address: 'Address',
                 block_height: int, block_hash: bytes,
                 tx_index: int, tx_hash: bytes,
                 iscore: int):
        super().__init__()

        self.msg_id = msg_id
        self.address: 'Address' = address
        self.block_height: int = block_height
        self.block_hash: bytes = block_hash
        self.tx_index: int = tx_index
        self.tx_hash: bytes = tx_hash
        self.iscore: int = iscore

    def __str__(self) -> str:
        return \
            f"CLAIM({self.msg_id}, " \
            f"{self.address}, " \
            f"{self.block_height}, {bytes_to_hex(self.block_hash)}, " \
            f"{self.tx_index}, {bytes_to_hex(self.tx_hash)}, " \
            f"{self.iscore})"

    @staticmethod
    def from_list(items: list) -> 'ClaimResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        address: 'Address' = MsgPackForIpc.decode(TypeTag.ADDRESS, payload[0])
        block_height: int = payload[1]
        block_hash: bytes = payload[2]
        tx_index: int = payload[3]
        tx_hash: bytes = payload[4]
        iscore: int = MsgPackForIpc.decode(TypeTag.INT, payload[5])

        return ClaimResponse(msg_id, address, block_height, block_hash, tx_index, tx_hash, iscore)


class CommitClaimRequest(Request):
    """Send the result of claimIScore tx to reward calculator
        No response for CommitClaimRequest
    """
    def __init__(self, success: bool, address: 'Address',
                 block_height: int, block_hash: bytes,
                 tx_index: int, tx_hash: bytes):
        super().__init__(MessageType.COMMIT_CLAIM)

        self.success = success
        self.address = address
        self.block_height = block_height
        self.block_hash = block_hash
        self.tx_index = tx_index
        self.tx_hash = tx_hash

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id, \
               (
                   self.success,
                   self.address.to_bytes_including_prefix(),
                   self.block_height,
                   self.block_hash,
                   self.tx_index,
                   self.tx_hash
               )

    def __str__(self) -> str:
        return \
            f"{self.msg_type.name}(" \
            f"{self.msg_id}, " \
            f"{self.success}, " \
            f"{self.address}, " \
            f"{self.block_height}, " \
            f"{bytes_to_hex(self.block_hash)}, " \
            f"{self.tx_index}, " \
            f"{bytes_to_hex(self.tx_hash)}" \
            f")"


class CommitClaimResponse(Response):
    MSG_TYPE = MessageType.COMMIT_CLAIM

    def __init__(self, msg_id: int):
        super().__init__()

        self.msg_id = msg_id

    def __str__(self) -> str:
        return f"COMMIT_CLAIM({self.msg_id})"

    @staticmethod
    def from_list(items: list) -> 'CommitClaimResponse':
        msg_id: int = items[1]

        return CommitClaimResponse(msg_id)


class QueryCalculateStatusRequest(Request):
    def __init__(self):
        super().__init__(MessageType.QUERY_CALCULATE_STATUS)

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id

    def __str__(self) -> str:
        return f"{self.msg_type.name}({self.msg_id})"


class QueryCalculateStatusResponse(Response):
    MSG_TYPE = MessageType.QUERY_CALCULATE_STATUS

    def __init__(self, msg_id: int, status: int, block_height: int):
        super().__init__()

        self.msg_id: int = msg_id
        self.status: int = status
        self.block_height: int = block_height

    def __str__(self) -> str:
        return f"QUERY_CALCULATE_STATUS_RESPONSE({self.msg_id}, {self.status}, {self.block_height})"

    @staticmethod
    def from_list(items: list) -> 'QueryCalculateStatusResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        status: int = payload[0]
        block_height: int = payload[1]

        return QueryCalculateStatusResponse(msg_id, status, block_height)


class QueryCalculateResultRequest(Request):

    def __init__(self, block_height):
        super().__init__(MessageType.QUERY_CALCULATE_RESULT)

        self.block_height = block_height

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id, self.block_height

    def __str__(self) -> str:
        return f"{self.msg_type.name}({self.msg_id},{self.block_height})"


class QueryCalculateResultResponse(Response):
    MSG_TYPE = MessageType.QUERY_CALCULATE_RESULT

    def __init__(self, msg_id: int, status: int, block_height: int, iscore: int, state_hash: bytes):
        super().__init__()

        self.msg_id = msg_id
        self.status = status
        self.block_height = block_height
        self.iscore = iscore
        self.state_hash = state_hash

    def __str__(self):
        return f"QUERY_CALCULATE_RESULT_RESPONSE({self.msg_id}, " \
            f"{self.status}, {self.block_height}, {self.iscore}, {bytes_to_hex(self.state_hash)})"

    @staticmethod
    def from_list(items: list) -> 'QueryCalculateResultResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        status: int = payload[0]
        block_hegiht: int = payload[1]
        iscore: int = MsgPackForIpc.decode(TypeTag.INT, payload[2])
        state_hash: bytes = payload[3]

        return QueryCalculateResultResponse(msg_id, status, block_hegiht, iscore, state_hash)


class QueryRequest(Request):
    """queryIScore
    """

    def __init__(self, address: 'Address'):
        super().__init__(MessageType.QUERY)

        self.address = address

    def __str__(self) -> str:
        return f"{self.msg_type.name}({self.msg_id}, {self.address})"

    def _to_list(self) -> tuple:
        return self.msg_type, \
               self.msg_id, \
               self.address.to_bytes_including_prefix()


class QueryResponse(Response):
    MSG_TYPE = MessageType.QUERY

    def __init__(self, msg_id: int,
                 address: 'Address', block_height: int, iscore: int):
        super().__init__()

        self.msg_id: int = msg_id
        self.address: 'Address' = address
        self.block_height: int = block_height
        self.iscore: int = iscore

    def __str__(self) -> str:
        return f"QUERY({self.msg_id}, {self.address}, {self.block_height}, {self.iscore})"

    @staticmethod
    def from_list(items: list) -> 'QueryResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        address: 'Address' = MsgPackForIpc.decode(TypeTag.ADDRESS, payload[0])
        iscore: int = MsgPackForIpc.decode(TypeTag.INT, payload[1])
        block_height: int = payload[2]

        return QueryResponse(msg_id, address, block_height, iscore)


class CalculateRequest(Request):
    def __init__(self, db_path: str, block_height: int):
        super().__init__(MessageType.CALCULATE)

        self.db_path = db_path
        self.block_height = block_height

    def __str__(self) -> str:
        return f"{self.msg_type.name}({self.msg_id}, {self.db_path}, {self.block_height})"

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id, (self.db_path, self.block_height)


class CalculateResponse(Response):
    MSG_TYPE = MessageType.CALCULATE

    def __init__(self, msg_id: int, status: int, block_height: int):
        super().__init__()

        self.msg_id: int = msg_id
        self.status: int = status
        self.block_height: int = block_height

    def __str__(self) -> str:
        return f"CALCULATE({self.msg_id}, {self.status}, {self.block_height})"

    @staticmethod
    def from_list(items: list) -> 'CalculateResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        status: int = payload[0]
        block_height: int = payload[1]

        return CalculateResponse(msg_id, status, block_height)


class CommitBlockRequest(Request):
    def __init__(self, success: bool, block_height: int, block_hash: bytes):
        super().__init__(MessageType.COMMIT_BLOCK)

        self.success = success
        self.block_height = block_height
        self.block_hash = block_hash

    def __str__(self):
        return f"{self.msg_type.name}({self.msg_id}, " \
            f"{self.success}, {self.block_height}, {bytes_to_hex(self.block_hash)})"

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id, (self.success, self.block_height, self.block_hash)


class CommitBlockResponse(Response):
    MSG_TYPE = MessageType.COMMIT_BLOCK

    def __init__(self, msg_id: int, success: bool, block_height: int, block_hash: bytes):
        super().__init__()

        self.msg_id: int = msg_id
        self.success: bool = success
        self.block_height: int = block_height
        self.block_hash: bytes = block_hash

    def __str__(self):
        return f"COMMIT_BLOCK({self.msg_id}, {self.success}, {self.block_height}, {bytes_to_hex(self.block_hash)})"

    @staticmethod
    def from_list(items: list) -> 'CommitBlockResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        success: bool = payload[0]
        block_height: int = payload[1]
        block_hash: bytes = payload[2]

        return CommitBlockResponse(msg_id, success, block_height, block_hash)


class InitRequest(Request):
    def __init__(self, block_height: int):
        super().__init__(MessageType.INIT)

        self.block_height = block_height

    def __str__(self):
        return f"{self.msg_type.name}({self.msg_id}, {self.block_height})"

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id, self.block_height


class InitResponse(Response):
    MSG_TYPE = MessageType.INIT

    def __init__(self, msg_id: int, success: bool, block_height: int):
        super().__init__()

        self.msg_id: int = msg_id
        self.success: bool = success
        self.block_height: int = block_height

    def __str__(self):
        return f"INIT({self.msg_id}, {self.success}, {self.block_height})"

    @staticmethod
    def from_list(items: list) -> 'InitResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        success: bool = payload[0]
        block_height: int = payload[1]

        return InitResponse(msg_id, success, block_height)


class RollbackRequest(Request):
    def __init__(self, block_height: int, block_hash: bytes):
        super().__init__(MessageType.ROLLBACK)

        self.block_height = block_height
        self.block_hash = block_hash

    def __str__(self):
        return f"{self.msg_type.name}({self.msg_id}, " \
               f"{self.block_height}, {bytes_to_hex(self.block_hash)})"

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id, (self.block_height, self.block_hash)


class RollbackResponse(Response):
    MSG_TYPE = MessageType.ROLLBACK

    def __init__(self, msg_id: int, success: bool, block_height: int, block_hash: bytes):
        super().__init__()

        self.msg_id: int = msg_id
        self.success: bool = success
        self.block_height: int = block_height
        self.block_hash: bytes = block_hash

    def __str__(self):
        return f"ROLLBACK({self.msg_id}, {self.success}, {self.block_height}, {bytes_to_hex(self.block_hash)})"

    @staticmethod
    def from_list(items: list) -> 'RollbackResponse':
        msg_id: int = items[1]
        payload: list = items[2]

        success: bool = payload[0]
        block_height: int = payload[1]
        block_hash: bytes = payload[2]

        return RollbackResponse(msg_id, success, block_height, block_hash)


class ReadyNotification(Response):
    MSG_TYPE = MessageType.READY

    def __init__(self, msg_id: int, version: int, block_height: int, block_hash: bytes):
        super().__init__()

        self.msg_id = msg_id
        self.version = version
        self.block_height = block_height
        self.block_hash = block_hash

    def __str__(self):
        return f"READY({self.msg_id}, {self.version}, {self.block_height}, {bytes_to_hex(self.block_hash)})"

    @staticmethod
    def from_list(items: list) -> 'ReadyNotification':
        msg_id: int = items[1]
        payload: list = items[2]

        version: int = payload[0]
        block_height: int = payload[1]
        block_hash: bytes = payload[2]

        return ReadyNotification(msg_id, version, block_height, block_hash)


class CalculateDoneNotification(Response):
    MSG_TYPE = MessageType.CALCULATE_DONE

    def __init__(self, msg_id: int, success: bool, block_height: int, iscore: int, state_hash: bytes):
        super().__init__()

        self.msg_id = msg_id
        self.success = success
        self.block_height = block_height
        self.iscore = iscore
        self.state_hash = state_hash

    def __str__(self):
        return f"CALCULATE_DONE({self.msg_id}, " \
            f"{self.success}, {self.block_height}, {self.iscore}, {bytes_to_hex(self.state_hash)})"

    @staticmethod
    def from_list(items: list) -> 'CalculateDoneNotification':
        msg_id: int = items[1]
        payload: list = items[2]

        success: bool = payload[0]
        block_hegiht: int = payload[1]
        iscore: int = MsgPackForIpc.decode(TypeTag.INT, payload[2])
        state_hash: bytes = payload[3]

        return CalculateDoneNotification(msg_id, success, block_hegiht, iscore, state_hash)


class NoneRequest(Request):
    """This request is used to stop ipc channel on iconservice stopping
    """
    def __init__(self):
        super().__init__(MessageType.NONE)

    def __str__(self):
        return f"NONE_REQUEST({self.msg_id})"

    def _to_list(self) -> tuple:
        return self.msg_type, self.msg_id


class NoneResponse(Response):
    MSG_TYPE = MessageType.NONE

    def __init__(self, msg_id: int):
        super().__init__()
        self.msg_id = msg_id

    def __str__(self):
        return f"NONE_RESPONSE({self.msg_id})"

    @staticmethod
    def from_list(items: list) -> 'NoneResponse':
        msg_id: int = items[1]
        return NoneResponse(msg_id)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional

from ...base.ComponentBase import StorageBase
from ...utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from ...iconscore.icon_score_context import IconScoreContext


class Storage(StorageBase):
    _REGULATOR_VARIABLE_KEY = b'regulator_variable'

    def get_regulator_variable(self, context: 'IconScoreContext'):
        regulator_variable: Optional[bytes] = self._db.get(context, self._REGULATOR_VARIABLE_KEY)
        if regulator_variable:
            return RegulatorVariable.from_bytes(regulator_variable)
        return RegulatorVariable(current_calc_period_issued_icx=0,
                                 prev_calc_period_issued_icx=-1,
                                 over_issued_iscore=0)

    def put_regulator_variable(self, context: 'IconScoreContext', rv: 'RegulatorVariable'):
        self._db.put(context, self._REGULATOR_VARIABLE_KEY, rv.to_bytes())


class RegulatorVariable:
    _VERSION = 0

    def __init__(self,
                 current_calc_period_issued_icx: int,
                 prev_calc_period_issued_icx: int,
                 over_issued_iscore: int):
        self.current_calc_period_issued_icx = current_calc_period_issued_icx
        self.prev_calc_period_issued_icx = prev_calc_period_issued_icx
        self.over_issued_iscore = over_issued_iscore

    @classmethod
    def from_bytes(cls, buf: bytes) -> 'RegulatorVariable':
        data: list = MsgPackForDB.loads(buf)
        version = data[0]

        return cls(*data[1:])

    def to_bytes(self):
        data: list = [
            self._VERSION,
            self.current_calc_period_issued_icx,
            self.prev_calc_period_issued_icx,
            self.over_issued_iscore
        ]
        return MsgPackForDB.dumps(data)

    def __str__(self):
        return f"Current calc period issued ICX: {self.current_calc_period_issued_icx} " \
               f"Prev calc period issued ICX: {self.prev_calc_period_issued_icx} " \
               f"Remain over issued I-SCORE: {self.over_issued_iscore} "
